import math
import unittest
import warnings

from igraph import Graph, InternalError, IN, OUT, ALL, TREE_IN
from math import inf, isnan


class SimplePropertiesTests(unittest.TestCase):
    gfull = Graph.Full(10)
    gempty = Graph(10)
    g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3)])
    gdir = Graph(
        4, [(0, 1), (0, 2), (1, 2), (2, 1), (0, 3), (1, 3), (3, 0)], directed=True
    )
    tree = Graph.Tree(14, 3)

    def testDensity(self):
        self.assertAlmostEqual(1.0, self.gfull.density(), places=5)
        self.assertAlmostEqual(0.0, self.gempty.density(), places=5)
        self.assertAlmostEqual(5 / 6, self.g.density(), places=5)
        self.assertAlmostEqual(1 / 2, self.g.density(True), places=5)
        self.assertAlmostEqual(7 / 12, self.gdir.density(), places=5)
        self.assertAlmostEqual(7 / 16, self.gdir.density(True), places=5)
        self.assertAlmostEqual(1 / 7, self.tree.density(), places=5)

    def testMeanDegree(self):
        self.assertEqual(9.0, self.gfull.mean_degree())
        self.assertEqual(0.0, self.gempty.mean_degree())
        self.assertEqual(2.5, self.g.mean_degree())
        self.assertEqual(7 / 4, self.gdir.mean_degree())
        self.assertAlmostEqual(13 / 7, self.tree.mean_degree(), places=5)

    def testDiameter(self):
        self.assertTrue(self.gfull.diameter() == 1)
        self.assertTrue(self.gempty.diameter(unconn=False) == inf)
        self.assertTrue(self.gempty.diameter(unconn=False) == inf)
        self.assertTrue(self.g.diameter() == 2)
        self.assertTrue(self.gdir.diameter(False) == 2)
        self.assertTrue(self.gdir.diameter() == 3)
        self.assertTrue(self.tree.diameter() == 5)

        s, t, d = self.tree.farthest_points()
        self.assertTrue((s == 13 or t == 13) and d == 5)
        self.assertTrue(self.gempty.farthest_points(unconn=False) == (None, None, inf))

        d = self.tree.get_diameter()
        self.assertTrue(d[0] == 13 or d[-1] == 13)

        weights = [1, 1, 1, 5, 1, 5, 1, 1, 1, 1, 1, 1, 5]
        self.assertTrue(self.tree.diameter(weights=weights) == 15)

        d = self.tree.farthest_points(weights=weights)
        self.assertTrue(d == (13, 6, 15) or d == (6, 13, 15))

    def testEccentricity(self):
        self.assertEqual(self.gfull.eccentricity(), [1] * self.gfull.vcount())
        self.assertEqual(self.gempty.eccentricity(), [0] * self.gempty.vcount())
        self.assertEqual(self.g.eccentricity(), [1, 1, 2, 2])
        self.assertEqual(self.gdir.eccentricity(), [1, 2, 3, 2])
        self.assertEqual(
            self.tree.eccentricity(), [3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5]
        )
        self.assertEqual(Graph().eccentricity(), [])

    def testWeightedEccentricity(self):
        self.assertEqual(
            self.gfull.eccentricity(weights=[2] * self.gfull.ecount()),
            [2] * self.gfull.vcount(),
        )
        self.assertEqual(
            self.gempty.eccentricity(weights=[]), [0] * self.gempty.vcount()
        )
        self.assertEqual(
            self.g.eccentricity(weights=range(1, self.g.ecount() + 1)), [4, 5, 6, 6]
        )
        self.assertEqual(Graph().eccentricity(), [])

    def testRadius(self):
        self.assertEqual(self.gfull.radius(), 1)
        self.assertEqual(self.gempty.radius(), 0)
        self.assertEqual(self.g.radius(), 1)
        self.assertEqual(self.gdir.radius(), 1)
        self.assertEqual(self.tree.radius(), 3)
        self.assertTrue(isnan(Graph().radius()))

    def testWeightedRadius(self):
        self.assertEqual(self.gfull.radius(weights=[2] * self.gfull.ecount()), 2)
        self.assertEqual(self.gempty.radius(weights=[]), 0)
        self.assertEqual(self.g.radius(weights=range(1, self.g.ecount() + 1)), 4)
        self.assertTrue(isnan(Graph().radius()))

    def testTransitivity(self):
        self.assertTrue(self.gfull.transitivity_undirected() == 1.0)
        self.assertTrue(self.tree.transitivity_undirected() == 0.0)
        self.assertTrue(self.g.transitivity_undirected() == 0.75)

    def testLocalTransitivity(self):
        self.assertTrue(
            self.gfull.transitivity_local_undirected() == [1.0] * self.gfull.vcount()
        )
        self.assertTrue(
            self.tree.transitivity_local_undirected(mode="zero")
            == [0.0] * self.tree.vcount()
        )

        transitivity = self.g.transitivity_local_undirected(mode="zero")
        self.assertAlmostEqual(2 / 3, transitivity[0], places=4)
        self.assertAlmostEqual(2 / 3, transitivity[1], places=4)
        self.assertEqual(1, transitivity[2])
        self.assertEqual(1, transitivity[3])

        g = Graph.Full(4) + 1 + [(0, 4)]
        g.es["weight"] = [1, 1, 1, 1, 1, 1, 5]
        self.assertAlmostEqual(
            g.transitivity_local_undirected(0, weights="weight"), 0.25, places=4
        )

    def testAvgLocalTransitivity(self):
        self.assertTrue(self.gfull.transitivity_avglocal_undirected() == 1.0)
        self.assertTrue(self.tree.transitivity_avglocal_undirected() == 0.0)
        self.assertAlmostEqual(
            self.g.transitivity_avglocal_undirected(), 5 / 6.0, places=4
        )

    def testModularity(self):
        g = Graph.Full(5) + Graph.Full(5)
        g.add_edges([(0, 5)])
        cl = [0] * 5 + [1] * 5
        self.assertAlmostEqual(g.modularity(cl), 0.4523, places=3)
        ws = [1] * 21
        self.assertAlmostEqual(g.modularity(cl, ws), 0.4523, places=3)
        ws = [2] * 21
        self.assertAlmostEqual(g.modularity(cl, ws), 0.4523, places=3)
        ws = [2] * 10 + [1] * 11
        self.assertAlmostEqual(g.modularity(cl, ws), 0.4157, places=3)
        self.assertRaises(InternalError, g.modularity, cl, ws[0:20])


class DegreeTests(unittest.TestCase):
    gfull = Graph.Full(10)
    gempty = Graph(10)
    g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (0, 0)])
    gdir = Graph(
        4, [(0, 1), (0, 2), (1, 2), (2, 1), (0, 3), (1, 3), (3, 0)], directed=True
    )
    tree = Graph.Tree(10, 3)

    def testKnn(self):
        knn, knnk = self.gfull.knn()
        self.assertTrue(knn == [9.0] * 10)
        self.assertAlmostEqual(knnk[8], 9.0, places=6)

        g = self.g.copy()
        g.simplify()

        knn, knnk = g.knn()
        diff = max(abs(a - b) for a, b in zip(knn, [7 / 3.0, 7 / 3.0, 3, 3]))
        self.assertAlmostEqual(diff, 0.0, places=6)
        self.assertEqual(len(knnk), 3)
        self.assertAlmostEqual(knnk[1], 3, places=6)
        self.assertAlmostEqual(knnk[2], 7 / 3.0, places=6)

    def testKnnNonSimple(self):
        knn, knnk = self.gfull.knn()
        self.assertTrue(knn == [9.0] * 10)
        self.assertAlmostEqual(knnk[8], 9.0, places=6)

        # knn works for non-simple graphs as well
        knn, knnk = self.g.knn()
        diff = max(abs(a - b) for a, b in zip(knn, [17 / 5.0, 3, 4, 4]))
        self.assertAlmostEqual(diff, 0.0, places=6)
        self.assertEqual(len(knnk), 5)
        self.assertAlmostEqual(knnk[1], 4, places=6)
        self.assertAlmostEqual(knnk[2], 3, places=6)
        self.assertAlmostEqual(knnk[4], 3.4, places=6)

    def testDegree(self):
        self.assertTrue(self.gfull.degree() == [9] * 10)
        self.assertTrue(self.gempty.degree() == [0] * 10)
        self.assertTrue(self.g.degree(loops=False) == [3, 3, 2, 2])
        self.assertTrue(self.g.degree() == [5, 3, 2, 2])
        self.assertTrue(self.gdir.degree(mode=IN) == [1, 2, 2, 2])
        self.assertTrue(self.gdir.degree(mode=OUT) == [3, 2, 1, 1])
        self.assertTrue(self.gdir.degree(mode=ALL) == [4, 4, 3, 3])
        vs = self.gdir.vs.select(0, 2)
        self.assertTrue(self.gdir.degree(vs, mode=ALL) == [4, 3])
        self.assertTrue(self.gdir.degree(self.gdir.vs[1], mode=ALL) == 4)

    def testMaxDegree(self):
        self.assertTrue(self.gfull.maxdegree() == 9)
        self.assertTrue(self.gempty.maxdegree() == 0)
        self.assertTrue(self.g.maxdegree() == 3)
        self.assertTrue(self.g.maxdegree(loops=True) == 5)
        self.assertTrue(self.g.maxdegree([1, 2], loops=True) == 3)
        self.assertTrue(self.gdir.maxdegree(mode=IN) == 2)
        self.assertTrue(self.gdir.maxdegree(mode=OUT) == 3)
        self.assertTrue(self.gdir.maxdegree(mode=ALL) == 4)

    def testStrength(self):
        # Turn off warnings about calling strength without weights
        import warnings

        warnings.filterwarnings(
            "ignore", "No edge weights for strength calculation", RuntimeWarning
        )

        # No weights
        self.assertTrue(self.gfull.strength() == [9] * 10)
        self.assertTrue(self.gempty.strength() == [0] * 10)
        self.assertTrue(self.g.degree(loops=False) == [3, 3, 2, 2])
        self.assertTrue(self.g.degree() == [5, 3, 2, 2])
        # With weights
        ws = [1, 2, 3, 4, 5, 6]
        self.assertTrue(self.g.strength(weights=ws, loops=False) == [7, 9, 5, 9])
        self.assertTrue(self.g.strength(weights=ws) == [19, 9, 5, 9])
        ws = [1, 2, 3, 4, 5, 6, 7]
        self.assertTrue(self.gdir.strength(mode=IN, weights=ws) == [7, 5, 5, 11])
        self.assertTrue(self.gdir.strength(mode=OUT, weights=ws) == [8, 9, 4, 7])
        self.assertTrue(self.gdir.strength(mode=ALL, weights=ws) == [15, 14, 9, 18])
        vs = self.gdir.vs.select(0, 2)
        self.assertTrue(self.gdir.strength(vs, mode=ALL, weights=ws) == [15, 9])
        self.assertTrue(self.gdir.strength(self.gdir.vs[1], mode=ALL, weights=ws) == 14)


class LocalTransitivityTests(unittest.TestCase):
    def testLocalTransitivityFull(self):
        trans = Graph.Full(10).transitivity_local_undirected()
        self.assertTrue(trans == [1.0] * 10)

    def testLocalTransitivityTree(self):
        trans = Graph.Tree(10, 3).transitivity_local_undirected()
        self.assertTrue(trans[0:3] == [0.0, 0.0, 0.0])

    def testLocalTransitivityHalf(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3)])
        trans = g.transitivity_local_undirected()
        trans = [round(x, 3) for x in trans]
        self.assertTrue(trans == [0.667, 0.667, 1.0, 1.0])

    def testLocalTransitivityPartial(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3)])
        trans = g.transitivity_local_undirected([1, 2])
        trans = [round(x, 3) for x in trans]
        self.assertTrue(trans == [0.667, 1.0])


class BiconnectedComponentTests(unittest.TestCase):
    g1 = Graph.Full(10)
    g2 = Graph(5, [(0, 1), (1, 2), (2, 3), (3, 4)])
    g3 = Graph(6, [(0, 1), (1, 2), (2, 3), (3, 0), (2, 4), (2, 5), (4, 5)])
    g4 = Graph.Full(2)
    g5 = Graph.Full(1)

    def testBiconnectedComponents(self):
        s = self.g1.biconnected_components()
        self.assertTrue(len(s) == 1 and s[0] == list(range(10)))
        s, ap = self.g1.biconnected_components(True)
        self.assertTrue(len(s) == 1 and s[0] == list(range(10)))

        s = self.g3.biconnected_components()
        self.assertTrue(len(s) == 2 and s[0] == [2, 4, 5] and s[1] == [0, 1, 2, 3])
        s, ap = self.g3.biconnected_components(True)
        self.assertTrue(
            len(s) == 2 and s[0] == [2, 4, 5] and s[1] == [0, 1, 2, 3] and ap == [2]
        )

    def testArticulationPoints(self):
        self.assertTrue(self.g1.articulation_points() == [])
        self.assertTrue(self.g2.cut_vertices() == [1, 2, 3])
        self.assertTrue(self.g3.articulation_points() == [2])
        self.assertTrue(self.g4.articulation_points() == [])

    def testIsBiconnected(self):
        self.assertTrue(self.g1.is_biconnected())
        self.assertFalse(self.g2.is_biconnected())
        self.assertFalse(self.g3.is_biconnected())
        self.assertTrue(self.g4.is_biconnected())
        self.assertFalse(self.g5.is_biconnected())


class CentralityTests(unittest.TestCase):
    def testBetweennessCentrality(self):
        g = Graph.Star(5)
        self.assertTrue(g.betweenness() == [6.0, 0.0, 0.0, 0.0, 0.0])

        g = Graph(5, [(0, 1), (0, 2), (0, 3), (1, 4)])
        self.assertTrue(g.betweenness() == [5.0, 3.0, 0.0, 0.0, 0.0])
        self.assertTrue(g.betweenness(cutoff=2) == [3.0, 1.0, 0.0, 0.0, 0.0])
        self.assertTrue(g.betweenness(cutoff=1) == [0.0, 0.0, 0.0, 0.0, 0.0])

        g = Graph.Lattice([3, 3], circular=False)
        self.assertTrue(
            g.betweenness(cutoff=2) == [0.5, 2.0, 0.5, 2.0, 4.0, 2.0, 0.5, 2.0, 0.5]
        )

        observed = g.betweenness(sources=[0, 8], targets=[0, 8])
        self.assertEqual(len(observed), g.vcount())
        for x, y in zip(
            observed, [0, 1 / 2, 1 / 6, 1 / 2, 2 / 3, 1 / 2, 1 / 6, 1 / 2, 0]
        ):
            self.assertAlmostEqual(x, y)
        self.assertRaises(
            ValueError, g.betweenness, cutoff=2, sources=[0, 8], targets=[0, 8]
        )

    def testEdgeBetweennessCentrality(self):
        g = Graph.Star(5)
        self.assertTrue(g.edge_betweenness() == [4.0, 4.0, 4.0, 4.0])

        g = Graph(5, [(0, 1), (0, 2), (0, 3), (1, 4)])
        self.assertTrue(g.edge_betweenness() == [6.0, 4.0, 4.0, 4.0])
        self.assertTrue(g.edge_betweenness(cutoff=2) == [4.0, 3.0, 3.0, 2.0])
        self.assertTrue(g.edge_betweenness(cutoff=1) == [1.0, 1.0, 1.0, 1.0])

        g = Graph.Ring(5)
        self.assertTrue(g.edge_betweenness() == [3.0, 3.0, 3.0, 3.0, 3.0])
        self.assertTrue(
            g.edge_betweenness(weights=[4, 1, 1, 1, 1]) == [0.5, 3.5, 5.5, 5.5, 3.5]
        )

        g = Graph.Lattice([3, 3], circular=False)
        observed = g.edge_betweenness(sources=[0, 8], targets=[0, 8])
        self.assertEqual(len(observed), g.ecount())
        for x, y in zip(
            observed,
            [
                1 / 2,
                1 / 2,
                1 / 6,
                1 / 3,
                1 / 6,
                1 / 3,
                1 / 6,
                1 / 3,
                1 / 3,
                1 / 2,
                1 / 6,
                1 / 2,
            ],
        ):
            self.assertAlmostEqual(x, y)
        self.assertRaises(
            ValueError, g.edge_betweenness, cutoff=2, sources=[0, 8], targets=[0, 8]
        )

    def testClosenessCentrality(self):
        g = Graph.Star(5)
        cl = g.closeness()
        cl2 = [1.0, 4 / 7.0, 4 / 7.0, 4 / 7.0, 4 / 7.0]
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        g = Graph.Star(5)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cl = g.closeness(cutoff=1.0)
        cl2 = [1.0, 1.0, 1.0, 1.0, 1.0]
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        weights = [1] * 4

        g = Graph.Star(5)
        cl = g.closeness(weights=weights)
        cl2 = [1.0, 0.57142, 0.57142, 0.57142, 0.57142]
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        g = Graph.Star(5)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cl = g.closeness(cutoff=1.0, weights=weights)
        cl2 = [1.0, 1.0, 1.0, 1.0, 1.0]
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        # Test for igraph/igraph:#1078
        g = Graph(
            [
                (0, 1),
                (0, 2),
                (0, 5),
                (0, 6),
                (0, 9),
                (1, 6),
                (1, 8),
                (2, 4),
                (2, 6),
                (2, 7),
                (2, 8),
                (3, 6),
                (4, 8),
                (5, 6),
                (5, 9),
                (6, 7),
                (6, 8),
                (7, 8),
                (7, 9),
                (8, 9),
            ]
        )
        weights = [
            0.69452,
            0.329886,
            0.131649,
            0.503269,
            0.472738,
            0.370933,
            0.23857,
            0.0354043,
            0.189015,
            0.355118,
            0.768335,
            0.893289,
            0.891709,
            0.494896,
            0.924684,
            0.432001,
            0.858159,
            0.246798,
            0.881304,
            0.64685,
        ]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cl = g.closeness(weights=weights)
        expected_cl = [
            1.63318,
            1.52014,
            2.03724,
            0.760158,
            1.91449,
            1.43224,
            1.91761,
            1.60198,
            1.3891,
            1.12829,
        ]
        for obs, exp in zip(cl, expected_cl):
            self.assertAlmostEqual(obs, exp, places=4)

    def testHarmonicCentrality(self):
        g = Graph.Star(5)
        cl = g.harmonic_centrality()
        cl2 = [1.0] + [(1.0 + 1 / 2 * 3) / 4] * 4
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        g = Graph.Star(5)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cl = g.harmonic_centrality(cutoff=1.0)
        cl2 = [1.0, 0.25, 0.25, 0.25, 0.25]
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        weights = [1] * 4

        g = Graph.Star(5)
        cl = g.harmonic_centrality(weights=weights)
        cl2 = [1.0] + [0.625] * 4
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

        g = Graph.Star(5)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cl = g.harmonic_centrality(cutoff=1.0, weights=weights)
        cl2 = [1.0, 0.25, 0.25, 0.25, 0.25]
        for idx in range(g.vcount()):
            self.assertAlmostEqual(cl[idx], cl2[idx], places=3)

    def testPageRank(self):
        g = Graph.Star(11)
        cent = g.pagerank()
        self.assertTrue(cent.index(max(cent)) == 0)
        self.assertAlmostEqual(max(cent), 0.4668, places=3)

    def testPersonalizedPageRank(self):
        g = Graph.Star(11)
        self.assertRaises(InternalError, g.personalized_pagerank, reset=[0] * 11)
        cent = g.personalized_pagerank(reset=[0, 10] + [0] * 9, damping=0.5)
        self.assertTrue(cent.index(max(cent)) == 1)
        self.assertAlmostEqual(cent[0], 0.3333, places=3)
        self.assertAlmostEqual(cent[1], 0.5166, places=3)
        self.assertAlmostEqual(cent[2], 0.0166, places=3)
        cent2 = g.personalized_pagerank(reset_vertices=g.vs[1], damping=0.5)
        self.assertTrue(max(abs(x - y) for x, y in zip(cent, cent2)) < 0.001)

    def testEigenvectorCentrality(self):
        g = Graph.Star(11)
        cent = g.evcent()
        self.assertTrue(cent.index(max(cent)) == 0)
        self.assertAlmostEqual(max(cent), 1.0, places=3)
        self.assertTrue(min(cent) >= 0)
        cent, ev = g.evcent(scale=False, return_eigenvalue=True)
        if cent[0] < 0:
            cent = [-x for x in cent]
        self.assertTrue(cent.index(max(cent)) == 0)
        self.assertAlmostEqual(cent[1] / cent[0], 0.3162, places=3)
        self.assertAlmostEqual(ev, 3.162, places=3)

    def testAuthorityScore(self):
        g = Graph.Tree(15, 2, TREE_IN)
        asc = g.authority_score()
        self.assertAlmostEqual(max(asc), 1.0, places=3)

        # Smoke testing
        g.authority_score(scale=False, return_eigenvalue=True)

    def testHubScore(self):
        g = Graph.Tree(15, 2, TREE_IN)
        hsc = g.hub_score()
        self.assertAlmostEqual(max(hsc), 1.0, places=3)

        # Smoke testing
        g.hub_score(scale=False, return_eigenvalue=True)

    def testCoreness(self):
        g = Graph.Full(4) + Graph(4) + [(0, 4), (1, 5), (2, 6), (3, 7)]
        self.assertEqual(g.coreness("all"), [3, 3, 3, 3, 1, 1, 1, 1])


class NeighborhoodTests(unittest.TestCase):
    def testNeighborhood(self):
        g = Graph.Ring(10, circular=False)
        self.assertTrue(
            list(map(sorted, g.neighborhood()))
            == [
                [0, 1],
                [0, 1, 2],
                [1, 2, 3],
                [2, 3, 4],
                [3, 4, 5],
                [4, 5, 6],
                [5, 6, 7],
                [6, 7, 8],
                [7, 8, 9],
                [8, 9],
            ]
        )
        self.assertTrue(
            list(map(sorted, g.neighborhood(order=3)))
            == [
                [0, 1, 2, 3],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4, 5],
                [0, 1, 2, 3, 4, 5, 6],
                [1, 2, 3, 4, 5, 6, 7],
                [2, 3, 4, 5, 6, 7, 8],
                [3, 4, 5, 6, 7, 8, 9],
                [4, 5, 6, 7, 8, 9],
                [5, 6, 7, 8, 9],
                [6, 7, 8, 9],
            ]
        )
        self.assertTrue(
            list(map(sorted, g.neighborhood(order=3, mindist=2)))
            == [
                [2, 3],
                [3, 4],
                [0, 4, 5],
                [0, 1, 5, 6],
                [1, 2, 6, 7],
                [2, 3, 7, 8],
                [3, 4, 8, 9],
                [4, 5, 9],
                [5, 6],
                [6, 7],
            ]
        )

    def testNeighborhoodSize(self):
        g = Graph.Ring(10, circular=False)
        self.assertTrue(g.neighborhood_size() == [2, 3, 3, 3, 3, 3, 3, 3, 3, 2])
        self.assertTrue(g.neighborhood_size(order=3) == [4, 5, 6, 7, 7, 7, 7, 6, 5, 4])
        self.assertTrue(
            g.neighborhood_size(order=3, mindist=2) == [2, 2, 3, 4, 4, 4, 4, 3, 2, 2]
        )


class MiscTests(unittest.TestCase):
    def assert_valid_maximum_cardinality_search_result(self, graph, alpha, alpham1):
        visited = []
        n = graph.vcount()
        not_visited = list(range(n))

        # Check if alpham1 is a valid visiting order
        for vertex in reversed(alpham1):
            neis = graph.neighbors(vertex)
            visited_neis = sum(1 for v in neis if v in visited)
            for other_vertex in not_visited:
                neis = graph.neighbors(other_vertex)
                other_visited_neis = sum(1 for v in neis if v in visited)
                self.assertTrue(other_visited_neis <= visited_neis)

            visited.append(vertex)
            not_visited.remove(vertex)

        # Check if alpha is the inverse of alpham1
        for index, vertex in enumerate(alpham1):
            self.assertEqual(alpha[vertex], index)

    def testBridges(self):
        g = Graph(5, [(0, 1), (1, 2), (2, 0), (0, 3), (3, 4)])
        self.assertEqual(g.bridges(), [3, 4])
        g = Graph(7, [(0, 1), (1, 2), (2, 0), (1, 6), (1, 3), (1, 4), (3, 5), (4, 5)])
        self.assertEqual(g.bridges(), [3])
        g = Graph(3, [(0, 1), (1, 2), (2, 3)])
        self.assertEqual(g.bridges(), [0, 1, 2])

    def testChordalCompletion(self):
        g = Graph()
        self.assertListEqual([], g.chordal_completion())

        g = Graph.Full(3)
        self.assertListEqual([], g.chordal_completion())

        g = Graph.Full(5)
        self.assertListEqual([], g.chordal_completion())

        g = Graph.Ring(4)
        cc = g.chordal_completion()
        self.assertEqual(len(cc), 1)
        g += cc
        self.assertTrue(g.is_chordal())
        self.assertListEqual([], g.chordal_completion())

        g = Graph.Ring(5)
        cc = g.chordal_completion()
        self.assertEqual(len(cc), 2)
        g += cc
        self.assertListEqual([], g.chordal_completion())

    def testChordalCompletionWithHints(self):
        g = Graph.Ring(4)
        alpha, _ = g.maximum_cardinality_search()
        cc = g.chordal_completion(alpha=alpha)
        self.assertEqual(len(cc), 1)
        g += cc
        self.assertTrue(g.is_chordal())
        self.assertListEqual([], g.chordal_completion())

        g = Graph.Ring(5)
        _, alpham1 = g.maximum_cardinality_search()
        cc = g.chordal_completion(alpham1=alpham1)
        self.assertEqual(len(cc), 2)
        g += cc
        self.assertListEqual([], g.chordal_completion())

    def testConstraint(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3)])
        self.assertTrue(isinstance(g.constraint(), list))  # TODO check more

    def testTopologicalSorting(self):
        g = Graph(5, [(0, 1), (0, 2), (1, 2), (1, 3), (2, 3)], directed=True)
        self.assertTrue(g.topological_sorting() == [0, 4, 1, 2, 3])
        self.assertTrue(g.topological_sorting(IN) == [3, 4, 2, 1, 0])
        g.to_undirected()
        self.assertRaises(InternalError, g.topological_sorting)

    def testIsTree(self):
        g = Graph()
        self.assertFalse(g.is_tree())

        g = Graph(directed=True)
        self.assertFalse(g.is_tree())

        g = Graph(1)
        self.assertTrue(g.is_tree())

        g = Graph(1, directed=True)
        self.assertTrue(
            g.is_tree() and g.is_tree("out") and g.is_tree("in") and g.is_tree("all")
        )

        g = Graph(5, [(0, 1), (1, 2), (1, 3), (3, 4)])
        self.assertTrue(g.is_tree())

        g = Graph(5, [(0, 1), (1, 2), (1, 3), (3, 4)], directed=True)
        self.assertTrue(g.is_tree())
        self.assertTrue(g.is_tree("out"))
        self.assertFalse(g.is_tree("in"))
        self.assertTrue(g.is_tree("all"))

        g = Graph(5, [(0, 1), (1, 2), (3, 1), (3, 4)], directed=True)
        self.assertFalse(g.is_tree())
        self.assertFalse(g.is_tree("in"))
        self.assertFalse(g.is_tree("out"))
        self.assertTrue(g.is_tree("all"))

        g = Graph(6, [(0, 4), (1, 5), (2, 1), (3, 1), (4, 3)], directed=True)
        self.assertFalse(g.is_tree())
        self.assertTrue(g.is_tree("in"))
        self.assertFalse(g.is_tree("out"))
        self.assertTrue(g.is_tree("all"))

        g = Graph.Ring(10)
        self.assertFalse(
            g.is_tree() or g.is_tree("in") or g.is_tree("out") or g.is_tree("all")
        )

    def testIsChordal(self):
        g = Graph()
        self.assertTrue(g.is_chordal())

        g = Graph.Full(3)
        self.assertTrue(g.is_chordal())

        g = Graph.Full(5)
        self.assertTrue(g.is_chordal())

        g = Graph.Ring(4)
        self.assertFalse(g.is_chordal())

        g = Graph.Ring(5)
        self.assertFalse(g.is_chordal())

    def testIsChordalWithHint(self):
        g = Graph()
        alpha, _ = g.maximum_cardinality_search()
        self.assertTrue(g.is_chordal(alpha=alpha))

        g = Graph.Full(3)
        alpha, _ = g.maximum_cardinality_search()
        self.assertTrue(g.is_chordal(alpha=alpha))

        g = Graph.Ring(5)
        alpha, _ = g.maximum_cardinality_search()
        self.assertFalse(g.is_chordal(alpha=alpha))

        g = Graph.Ring(4)
        _, alpham1 = g.maximum_cardinality_search()
        self.assertFalse(g.is_chordal(alpham1=alpham1))

        g = Graph.Full(5)
        _, alpham1 = g.maximum_cardinality_search()
        self.assertTrue(g.is_chordal(alpham1=alpham1))

    def testLineGraph(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3)])
        el = g.linegraph().get_edgelist()
        el.sort()
        self.assertTrue(
            el == [(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3), (2, 4), (3, 4)]
        )

        g = Graph(4, [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3)], directed=True)
        el = g.linegraph().get_edgelist()
        el.sort()
        self.assertTrue(el == [(0, 2), (0, 4)])

    def testMaximumCardinalitySearch(self):
        g = Graph()
        alpha, alpham1 = g.maximum_cardinality_search()
        self.assertListEqual([], alpha)
        self.assertListEqual([], alpham1)

        g = Graph.Famous("petersen")
        alpha, alpham1 = g.maximum_cardinality_search()

        self.assert_valid_maximum_cardinality_search_result(g, alpha, alpham1)

        g = Graph.GRG(100, 0.2)
        alpha, alpham1 = g.maximum_cardinality_search()

        self.assert_valid_maximum_cardinality_search_result(g, alpha, alpham1)


class PathTests(unittest.TestCase):
    def testDistances(self):
        g = Graph(
            10,
            [
                (0, 1),
                (0, 2),
                (0, 3),
                (1, 2),
                (1, 4),
                (1, 5),
                (2, 3),
                (2, 6),
                (3, 2),
                (3, 6),
                (4, 5),
                (4, 7),
                (5, 6),
                (5, 8),
                (5, 9),
                (7, 5),
                (7, 8),
                (8, 9),
                (5, 2),
                (2, 1),
            ],
            directed=True,
        )
        ws = [0, 2, 1, 0, 5, 2, 1, 1, 0, 2, 2, 8, 1, 1, 3, 1, 1, 4, 2, 1]
        g.es["weight"] = ws
        expected = [
            [0, 0, 0, 1, 5, 2, 1, 13, 3, 5],
            [inf, 0, 0, 1, 5, 2, 1, 13, 3, 5],
            [inf, 1, 0, 1, 6, 3, 1, 14, 4, 6],
            [inf, 1, 0, 0, 6, 3, 1, 14, 4, 6],
            [inf, 5, 4, 5, 0, 2, 3, 8, 3, 5],
            [inf, 3, 2, 3, 8, 0, 1, 16, 1, 3],
            [inf, inf, inf, inf, inf, inf, 0, inf, inf, inf],
            [inf, 4, 3, 4, 9, 1, 2, 0, 1, 4],
            [inf, inf, inf, inf, inf, inf, inf, inf, 0, 4],
            [inf, inf, inf, inf, inf, inf, inf, inf, inf, 0],
        ]
        self.assertTrue(g.distances(weights=ws) == expected)
        self.assertTrue(g.distances(weights="weight") == expected)
        self.assertTrue(
            g.distances(weights="weight", target=[2, 3])
            == [row[2:4] for row in expected]
        )
        self.assertTrue(
            g.distances(weights="weight", target=[2, 3], algorithm="bellman_ford")
            == [row[2:4] for row in expected]
        )
        self.assertTrue(
            g.distances(weights="weight", target=[2, 3], algorithm="johnson")
            == [row[2:4] for row in expected]
        )
        self.assertRaises(
            ValueError,
            g.distances,
            weights="weight",
            target=[2, 3],
            algorithm="johnson",
            mode="in",
        )

    def testGetShortestPath(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 3), (3, 2), (2, 1)], directed=True)
        self.assertEqual([0, 1, 3], g.get_shortest_path(0, 3))
        self.assertEqual([0, 1, 3], g.get_shortest_path(0, 3, output="vpath"))
        self.assertEqual([0, 2], g.get_shortest_path(0, 3, output="epath"))
        self.assertRaises(ValueError, g.get_shortest_path, 0, 3, output="x")
        self.assertRaises(TypeError, g.get_shortest_path, 0)

    def testGetShortestPathManualAlgorithmSelection(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 3), (3, 2), (2, 1)], directed=True)
        g.es["weight"] = [1] * g.ecount()

        self.assertEqual([0, 1, 3], g.get_shortest_path(0, 3, algorithm="bellman_ford"))
        self.assertRaises(ValueError, g.get_shortest_path, 0, 3, algorithm="johnson")
        self.assertRaises(
            ValueError, g.get_shortest_path, 0, 3, algorithm="johnson", mode="in"
        )

    def testGetShortestPaths(self):
        g = Graph(4, [(0, 1), (0, 2), (1, 3), (3, 2), (2, 1)], directed=True)
        sps = g.get_shortest_paths(0)
        expected = [[0], [0, 1], [0, 2], [0, 1, 3]]
        self.assertTrue(sps == expected)
        sps = g.get_shortest_paths(0, output="vpath")
        expected = [[0], [0, 1], [0, 2], [0, 1, 3]]
        self.assertTrue(sps == expected)
        sps = g.get_shortest_paths(0, output="epath")
        expected = [[], [0], [1], [0, 2]]
        self.assertTrue(sps == expected)
        self.assertRaises(ValueError, g.get_shortest_paths, 0, output="x")

    def testGetAllShortestPaths(self):
        g = Graph(4, [(0, 1), (1, 2), (1, 3), (2, 4), (3, 4), (4, 5)], directed=True)

        sps = sorted(g.get_all_shortest_paths(0, 0))
        expected = [[0]]
        self.assertEqual(expected, sps)

        sps = sorted(g.get_all_shortest_paths(0, 5))
        expected = [[0, 1, 2, 4, 5], [0, 1, 3, 4, 5]]
        self.assertEqual(expected, sps)

        sps = sorted(g.get_all_shortest_paths(1, 4))
        expected = [[1, 2, 4], [1, 3, 4]]
        self.assertEqual(expected, sps)

        g = Graph.Lattice([5, 5], circular=False)

        sps = sorted(g.get_all_shortest_paths(0, 12))
        expected = [
            [0, 1, 2, 7, 12],
            [0, 1, 6, 7, 12],
            [0, 1, 6, 11, 12],
            [0, 5, 6, 7, 12],
            [0, 5, 6, 11, 12],
            [0, 5, 10, 11, 12],
        ]
        self.assertEqual(expected, sps)

        g = Graph.Lattice([100, 100], circular=False)
        sps = sorted(g.get_all_shortest_paths(0, 202))
        expected = [
            [0, 1, 2, 102, 202],
            [0, 1, 101, 102, 202],
            [0, 1, 101, 201, 202],
            [0, 100, 101, 102, 202],
            [0, 100, 101, 201, 202],
            [0, 100, 200, 201, 202],
        ]
        self.assertEqual(expected, sps)

        g = Graph.Lattice([100, 100], circular=False)
        sps = sorted(g.get_all_shortest_paths(0, [0, 202]))
        self.assertEqual([[0]] + expected, sps)

        g = Graph([(0, 1), (1, 2), (0, 2)])
        g.es["weight"] = [0.5, 0.5, 1]
        sps = sorted(g.get_all_shortest_paths(0, weights="weight"))
        self.assertEqual([[0], [0, 1], [0, 1, 2], [0, 2]], sps)

        g = Graph.Lattice([4, 4], circular=False)
        g.es["weight"] = 1
        g.es[2, 8]["weight"] = 100
        sps = sorted(g.get_all_shortest_paths(0, [3, 12, 15], weights="weight"))
        self.assertEqual(20, len(sps))
        self.assertEqual(4, sum(1 for path in sps if path[-1] == 3))
        self.assertEqual(4, sum(1 for path in sps if path[-1] == 12))
        self.assertEqual(12, sum(1 for path in sps if path[-1] == 15))

    def testGetKShortestPaths(self):
        g = Graph(4, [(0, 1), (1, 2), (1, 3), (2, 4), (3, 4), (4, 5)], directed=True)

        sps = sorted(g.get_k_shortest_paths(0, 0))
        expected = [[0]]
        self.assertEqual(expected, sps)

        sps = sorted(g.get_k_shortest_paths(0, 5, 2))
        expected = [[0, 1, 2, 4, 5], [0, 1, 3, 4, 5]]
        self.assertEqual(expected, sps)

        sps = sorted(g.get_k_shortest_paths(1, 4, 2))
        expected = [[1, 2, 4], [1, 3, 4]]
        self.assertEqual(expected, sps)

        g = Graph.Lattice([5, 5], circular=False)

        sps = sorted(g.get_k_shortest_paths(0, 12, 6))
        expected = [
            [0, 1, 2, 7, 12],
            [0, 1, 6, 7, 12],
            [0, 1, 6, 11, 12],
            [0, 5, 6, 7, 12],
            [0, 5, 6, 11, 12],
            [0, 5, 10, 11, 12],
        ]
        self.assertEqual(expected, sps)

        g = Graph.Lattice([100, 100], circular=False)
        sps = sorted(g.get_k_shortest_paths(0, 202, 6))
        expected = [
            [0, 1, 2, 102, 202],
            [0, 1, 101, 102, 202],
            [0, 1, 101, 201, 202],
            [0, 100, 101, 102, 202],
            [0, 100, 101, 201, 202],
            [0, 100, 200, 201, 202],
        ]
        self.assertEqual(expected, sps)

        g = Graph([(0, 1), (1, 2), (0, 2)])
        g.es["weight"] = [0.5, 0.5, 1]
        sps = sorted(g.get_k_shortest_paths(0, 2, 2, weights="weight"))
        self.assertEqual(sorted([[0, 2], [0, 1, 2]]), sorted(sps))

    def testGetAllSimplePaths(self):
        g = Graph.Ring(20)
        sps = sorted(g.get_all_simple_paths(0, 10))
        self.assertEqual(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [0, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10],
            ],
            sps,
        )

        g = Graph.Ring(20, directed=True)
        sps = sorted(g.get_all_simple_paths(0, 10))
        self.assertEqual([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]], sps)
        sps = sorted(g.get_all_simple_paths(0, 10, mode="in"))
        self.assertEqual([[0, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10]], sps)
        sps = sorted(g.get_all_simple_paths(0, 10, mode="all"))
        self.assertEqual(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [0, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10],
            ],
            sps,
        )

        g = Graph.Lattice([4, 4], circular=False)
        g = Graph([(min(u, v), max(u, v)) for u, v in g.get_edgelist()], directed=True)
        sps = sorted(g.get_all_simple_paths(0, 15))
        self.assertEqual(20, len(sps))
        for path in sps:
            self.assertEqual(0, path[0])
            self.assertEqual(15, path[-1])
            curr = path[0]
            for next in path[1:]:
                self.assertTrue(g.are_adjacent(curr, next))
                curr = next

    def testPathLengthHist(self):
        g = Graph.Tree(15, 2)
        h = g.path_length_hist()
        self.assertTrue(h.unconnected == 0)
        self.assertTrue(
            [(int(value), x) for value, _, x in h.bins()]
            == [(1, 14), (2, 19), (3, 20), (4, 20), (5, 16), (6, 16)]
        )
        g = Graph.Full(5) + Graph.Full(4)
        h = g.path_length_hist()
        self.assertTrue(h.unconnected == 20)
        g.to_directed()
        h = g.path_length_hist()
        self.assertTrue(h.unconnected == 40)
        h = g.path_length_hist(False)
        self.assertTrue(h.unconnected == 20)

    def testGetShortestPathsAStar(self):
        n = 4
        g = Graph.Lattice((n, n), circular=False)
        xs, ys = list(range(n)) * n, sum(([i] * n for i in range(n)), [])
        g.vs["coord"] = list(zip(xs, ys))

        def heuristics(graph, u, v):
            ux, uy = graph.vs[u]["coord"]
            vx, vy = graph.vs[v]["coord"]
            return ((ux - vx) ** 2 + (uy - vy) ** 2) ** 0.5

        self.assertEqual(
            [0, 1, 5, 6, 10, 11], g.get_shortest_path_astar(0, 11, heuristics)
        )


class DominatorTests(unittest.TestCase):
    def compareDomTrees(self, alist, blist):
        """
        Required due to NaN use for isolated nodes
        """
        if len(alist) != len(blist):
            return False
        for _i, (a, b) in enumerate(zip(alist, blist)):
            if math.isnan(a) and math.isnan(b):
                continue
            elif a == b:
                continue
            else:
                return False
        return True

    def testDominators(self):
        # examples taken from igraph's examples/simple/dominator_tree.out

        # initial
        g = Graph(
            13,
            [
                (0, 1),
                (0, 7),
                (0, 10),
                (1, 2),
                (1, 5),
                (2, 3),
                (3, 4),
                (4, 3),
                (4, 0),
                (5, 3),
                (5, 6),
                (6, 3),
                (7, 8),
                (7, 10),
                (7, 11),
                (8, 9),
                (9, 4),
                (9, 8),
                (10, 11),
                (11, 12),
                (12, 9),
            ],
            directed=True,
        )
        s = [-1, 0, 1, 0, 0, 1, 5, 0, 0, 0, 0, 0, 11]
        r = g.dominator(0)
        self.assertTrue(self.compareDomTrees(s, r))

        # flipped edges
        g = Graph(
            13,
            [
                (1, 0),
                (2, 0),
                (3, 0),
                (4, 1),
                (1, 2),
                (4, 2),
                (5, 2),
                (6, 3),
                (7, 3),
                (12, 4),
                (8, 5),
                (9, 6),
                (9, 7),
                (10, 7),
                (5, 8),
                (11, 8),
                (11, 9),
                (9, 10),
                (9, 11),
                (0, 11),
                (8, 12),
            ],
            directed=True,
        )
        s = [-1, 0, 0, 0, 0, 0, 3, 3, 0, 0, 7, 0, 4]
        r = g.dominator(0, mode=IN)
        self.assertTrue(self.compareDomTrees(s, r))

        # disconnected components
        g = Graph(
            20,
            [
                (0, 1),
                (0, 2),
                (0, 3),
                (1, 4),
                (2, 1),
                (2, 4),
                (2, 8),
                (3, 9),
                (3, 10),
                (4, 15),
                (8, 11),
                (9, 12),
                (10, 12),
                (10, 13),
                (11, 8),
                (11, 14),
                (12, 14),
                (13, 12),
                (14, 12),
                (14, 0),
                (15, 11),
            ],
            directed=True,
        )
        s = [
            -1,
            0,
            0,
            0,
            0,
            float("nan"),
            float("nan"),
            float("nan"),
            0,
            3,
            3,
            0,
            0,
            10,
            0,
            4,
            float("nan"),
            float("nan"),
            float("nan"),
            float("nan"),
        ]
        r = g.dominator(0, mode=OUT)
        self.assertTrue(self.compareDomTrees(s, r))


def suite():
    simple_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        SimplePropertiesTests
    )
    degree_suite = unittest.defaultTestLoader.loadTestsFromTestCase(DegreeTests)
    local_transitivity_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        LocalTransitivityTests
    )
    biconnected_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        BiconnectedComponentTests
    )
    centrality_suite = unittest.defaultTestLoader.loadTestsFromTestCase(CentralityTests)
    neighborhood_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        NeighborhoodTests
    )
    path_suite = unittest.defaultTestLoader.loadTestsFromTestCase(PathTests)
    misc_suite = unittest.defaultTestLoader.loadTestsFromTestCase(MiscTests)
    dominator_suite = unittest.defaultTestLoader.loadTestsFromTestCase(DominatorTests)
    return unittest.TestSuite(
        [
            simple_suite,
            degree_suite,
            local_transitivity_suite,
            biconnected_suite,
            centrality_suite,
            neighborhood_suite,
            path_suite,
            misc_suite,
            dominator_suite,
        ]
    )


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



import unittest
from random import shuffle

from igraph import Graph


def node_compat(g1, g2, v1, v2):
    """Node compatibility function for isomorphism tests"""
    return g1.vs[v1]["color"] == g2.vs[v2]["color"]


def edge_compat(g1, g2, e1, e2):
    """Edge compatibility function for isomorphism tests"""
    return g1.es[e1]["color"] == g2.es[e2]["color"]


class IsomorphismTests(unittest.TestCase):
    def testIsomorphic(self):
        g1 = Graph(
            8,
            [
                (0, 4),
                (0, 5),
                (0, 6),
                (1, 4),
                (1, 5),
                (1, 7),
                (2, 4),
                (2, 6),
                (2, 7),
                (3, 5),
                (3, 6),
                (3, 7),
            ],
        )
        g2 = Graph(
            8,
            [
                (0, 1),
                (0, 3),
                (0, 4),
                (2, 3),
                (2, 1),
                (2, 6),
                (5, 1),
                (5, 4),
                (5, 6),
                (7, 3),
                (7, 6),
                (7, 4),
            ],
        )

        # Test the isomorphism of g1 and g2
        self.assertTrue(g1.isomorphic(g2))
        self.assertTrue(
            g2.isomorphic_vf2(g1, return_mapping_21=True)
            == (True, None, [0, 2, 5, 7, 1, 3, 4, 6])
        )
        self.assertTrue(
            g2.isomorphic_bliss(g1, return_mapping_21=True, sh1="fl")
            == (True, None, [0, 2, 5, 7, 1, 3, 4, 6])
        )
        self.assertRaises(ValueError, g2.isomorphic_bliss, g1, sh2="nonexistent")

        # Test the automorphy of g1
        self.assertTrue(g1.isomorphic())
        self.assertTrue(
            g1.isomorphic_vf2(return_mapping_21=True)
            == (True, None, [0, 1, 2, 3, 4, 5, 6, 7])
        )

        # Test VF2 with colors
        self.assertTrue(
            g1.isomorphic_vf2(
                g2, color1=[0, 1, 0, 1, 0, 1, 0, 1], color2=[0, 0, 1, 1, 0, 0, 1, 1]
            )
        )
        g1.vs["color"] = [0, 1, 0, 1, 0, 1, 0, 1]
        g2.vs["color"] = [0, 0, 1, 1, 0, 1, 1, 0]
        self.assertTrue(not g1.isomorphic_vf2(g2, "color", "color"))

        # Test bliss with colors
        self.assertTrue(
            g1.isomorphic_bliss(
                g2, color1=[0, 0, 0, 0, 0, 0, 0, 0], color2=[0, 0, 0, 0, 0, 0, 0, 0]
            )
        )

        self.assertTrue(
            g1.isomorphic_bliss(
                g2, color1=[1, 0, 2, 0, 0, 0, 0, 0], color2=[1, 0, 2, 0, 0, 0, 0, 0]
            )
        )

        self.assertTrue(
            g1.isomorphic_bliss(
                g2, color1=[0, 1, 0, 1, 0, 1, 0, 1], color2=[0, 0, 1, 1, 0, 0, 1, 1]
            )
        )

        # Test VF2 with vertex and edge colors
        self.assertTrue(
            g1.isomorphic_vf2(
                g2, color1=[0, 1, 0, 1, 0, 1, 0, 1], color2=[0, 0, 1, 1, 0, 0, 1, 1]
            )
        )
        g1.es["color"] = list(range(12))
        g2.es["color"] = [0] * 6 + [1] * 6
        self.assertTrue(not g1.isomorphic_vf2(g2, "color", "color", "color", "color"))

        # Test VF2 with node compatibility function
        g2.vs["color"] = [0, 0, 1, 1, 0, 0, 1, 1]
        self.assertTrue(g1.isomorphic_vf2(g2, node_compat_fn=node_compat))
        g2.vs["color"] = [0, 0, 1, 1, 0, 1, 1, 0]
        self.assertTrue(not g1.isomorphic_vf2(g2, node_compat_fn=node_compat))

        # Test VF2 with node edge compatibility function
        g2.vs["color"] = [0, 0, 1, 1, 0, 0, 1, 1]
        g1.es["color"] = list(range(12))
        g2.es["color"] = [0] * 6 + [1] * 6
        self.assertTrue(
            not g1.isomorphic_vf2(
                g2, node_compat_fn=node_compat, edge_compat_fn=edge_compat
            )
        )

    def testIsomorphicCallback(self):
        maps = []

        def callback(g1, g2, map1, map2):
            maps.append(map1)
            return True

        # Test VF2 callback
        g = Graph(6, [(0, 1), (2, 3), (4, 5), (0, 2), (2, 4), (1, 3), (3, 5)])
        g.isomorphic_vf2(g, callback=callback)
        expected_maps = [
            [0, 1, 2, 3, 4, 5],
            [1, 0, 3, 2, 5, 4],
            [4, 5, 2, 3, 0, 1],
            [5, 4, 3, 2, 1, 0],
        ]
        self.assertTrue(sorted(maps) == expected_maps)

        maps[:] = []
        g3 = Graph.Full(4)
        g3.vs["color"] = [0, 1, 1, 0]
        g3.isomorphic_vf2(callback=callback, color1="color", color2="color")
        expected_maps = [[0, 1, 2, 3], [0, 2, 1, 3], [3, 1, 2, 0], [3, 2, 1, 0]]
        self.assertTrue(sorted(maps) == expected_maps)

    def testCountIsomorphisms(self):
        g = Graph.Full(4)
        self.assertTrue(g.count_automorphisms_vf2() == 24)
        g = Graph(6, [(0, 1), (2, 3), (4, 5), (0, 2), (2, 4), (1, 3), (3, 5)])
        self.assertTrue(g.count_automorphisms_vf2() == 4)

        # Some more tests with colors
        g3 = Graph.Full(4)
        g3.vs["color"] = [0, 1, 1, 0]
        self.assertTrue(g3.count_isomorphisms_vf2() == 24)
        self.assertTrue(g3.count_isomorphisms_vf2(color1="color", color2="color") == 4)
        self.assertTrue(
            g3.count_isomorphisms_vf2(color1=[0, 1, 2, 0], color2=(0, 1, 2, 0)) == 2
        )
        self.assertTrue(
            g3.count_isomorphisms_vf2(
                edge_color1=[0, 1, 0, 0, 0, 1], edge_color2=[0, 1, 0, 0, 0, 1]
            )
            == 2
        )

        # Test VF2 with node/edge compatibility function
        g3.vs["color"] = [0, 1, 1, 0]
        self.assertTrue(g3.count_isomorphisms_vf2(node_compat_fn=node_compat) == 4)
        g3.vs["color"] = [0, 1, 2, 0]
        self.assertTrue(g3.count_isomorphisms_vf2(node_compat_fn=node_compat) == 2)
        g3.es["color"] = [0, 1, 0, 0, 0, 1]
        self.assertTrue(g3.count_isomorphisms_vf2(edge_compat_fn=edge_compat) == 2)

    def testGetIsomorphisms(self):
        g = Graph(6, [(0, 1), (2, 3), (4, 5), (0, 2), (2, 4), (1, 3), (3, 5)])
        maps = g.get_automorphisms_vf2()
        expected_maps = [
            [0, 1, 2, 3, 4, 5],
            [1, 0, 3, 2, 5, 4],
            [4, 5, 2, 3, 0, 1],
            [5, 4, 3, 2, 1, 0],
        ]
        self.assertTrue(maps == expected_maps)

        g3 = Graph.Full(4)
        g3.vs["color"] = [0, 1, 1, 0]
        expected_maps = [[0, 1, 2, 3], [0, 2, 1, 3], [3, 1, 2, 0], [3, 2, 1, 0]]
        self.assertTrue(
            sorted(g3.get_automorphisms_vf2(color="color")) == expected_maps
        )


class SubisomorphismTests(unittest.TestCase):
    def testSubisomorphicLAD(self):
        g = Graph.Lattice([3, 3], circular=False)
        g2 = Graph([(0, 1), (1, 2), (1, 3)])
        g3 = g + [(0, 4), (2, 4), (6, 4), (8, 4), (3, 1), (1, 5), (5, 7), (7, 3)]

        self.assertTrue(g.subisomorphic_lad(g2))
        self.assertFalse(g2.subisomorphic_lad(g))

        # Test 'induced'
        self.assertFalse(g3.subisomorphic_lad(g, induced=True))
        self.assertTrue(g3.subisomorphic_lad(g, induced=False))
        self.assertTrue(g3.subisomorphic_lad(g))
        self.assertTrue(g3.subisomorphic_lad(g2, induced=True))
        self.assertTrue(g3.subisomorphic_lad(g2, induced=False))
        self.assertTrue(g3.subisomorphic_lad(g2))

        # Test with limited vertex matching
        domains = [
            [4],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
        ]
        self.assertTrue(g.subisomorphic_lad(g2, domains=domains))
        domains = [
            [],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
        ]
        self.assertTrue(not g.subisomorphic_lad(g2, domains=domains))

        # Corner cases
        empty = Graph()
        self.assertTrue(g.subisomorphic_lad(empty))
        self.assertTrue(empty.subisomorphic_lad(empty))

    def testGetSubisomorphismsLAD(self):
        g = Graph.Lattice([3, 3], circular=False)
        g2 = Graph([(0, 1), (1, 2), (2, 3), (3, 0)])
        g3 = g + [(0, 4), (2, 4), (6, 4), (8, 4), (3, 1), (1, 5), (5, 7), (7, 3)]

        all_subiso = "0143 0341 1034 1254 1430 1452 2145 2541 3014 3410 3476 \
        3674 4103 4125 4301 4367 4521 4587 4763 4785 5214 5412 5478 5874 6347 \
        6743 7436 7458 7634 7854 8547 8745"
        all_subiso = sorted([int(x) for x in item] for item in all_subiso.split())

        self.assertEqual(all_subiso, sorted(g.get_subisomorphisms_lad(g2)))
        self.assertEqual([], sorted(g2.get_subisomorphisms_lad(g)))

        # Test 'induced'
        induced_subiso = "1375 1573 3751 5731 7513 7315 5137 3157"
        induced_subiso = sorted(
            [int(x) for x in item] for item in induced_subiso.split()
        )
        self.assertEqual(
            induced_subiso, sorted(g3.get_subisomorphisms_lad(g2, induced=True))
        )
        self.assertEqual([], g3.get_subisomorphisms_lad(g, induced=True))

        # Test with limited vertex matching
        limited_subiso = [iso for iso in all_subiso if iso[0] == 4]
        domains = [
            [4],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
        ]
        self.assertEqual(
            limited_subiso, sorted(g.get_subisomorphisms_lad(g2, domains=domains))
        )
        domains = [
            [],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
            [0, 1, 2, 3, 5, 6, 7, 8],
        ]
        self.assertEqual([], sorted(g.get_subisomorphisms_lad(g2, domains=domains)))

        # Corner cases
        empty = Graph()
        self.assertEqual([[]], g.get_subisomorphisms_lad(empty))
        self.assertEqual([[]], empty.get_subisomorphisms_lad(empty))

    def testSubisomorphicVF2(self):
        g = Graph.Lattice([3, 3], circular=False)
        g2 = Graph([(0, 1), (1, 2), (1, 3)])
        self.assertTrue(g.subisomorphic_vf2(g2))
        self.assertTrue(not g2.subisomorphic_vf2(g))

        # Test with vertex colors
        g.vs["color"] = [0, 0, 0, 0, 1, 0, 0, 0, 0]
        g2.vs["color"] = [1, 0, 0, 0]
        self.assertTrue(g.subisomorphic_vf2(g2, node_compat_fn=node_compat))
        g2.vs["color"] = [2, 0, 0, 0]
        self.assertTrue(not g.subisomorphic_vf2(g2, node_compat_fn=node_compat))

        # Test with edge colors
        g.es["color"] = [1] + [0] * (g.ecount() - 1)
        g2.es["color"] = [1] + [0] * (g2.ecount() - 1)
        self.assertTrue(g.subisomorphic_vf2(g2, edge_compat_fn=edge_compat))
        g2.es[0]["color"] = [2]
        self.assertTrue(not g.subisomorphic_vf2(g2, node_compat_fn=node_compat))

    def testCountSubisomorphisms(self):
        g = Graph.Lattice([3, 3], circular=False)
        g2 = Graph.Lattice([2, 2], circular=False)
        self.assertTrue(g.count_subisomorphisms_vf2(g2) == 4 * 4 * 2)
        self.assertTrue(g2.count_subisomorphisms_vf2(g) == 0)

        # Test with vertex colors
        g.vs["color"] = [0, 0, 0, 0, 1, 0, 0, 0, 0]
        g2.vs["color"] = [1, 0, 0, 0]
        self.assertTrue(g.count_subisomorphisms_vf2(g2, "color", "color") == 4 * 2)
        self.assertTrue(
            g.count_subisomorphisms_vf2(g2, node_compat_fn=node_compat) == 4 * 2
        )

        # Test with edge colors
        g.es["color"] = [1] + [0] * (g.ecount() - 1)
        g2.es["color"] = [1] + [0] * (g2.ecount() - 1)
        self.assertTrue(
            g.count_subisomorphisms_vf2(g2, edge_color1="color", edge_color2="color")
            == 2
        )
        self.assertTrue(
            g.count_subisomorphisms_vf2(g2, edge_compat_fn=edge_compat) == 2
        )


class PermutationTests(unittest.TestCase):
    def testCanonicalPermutation(self):
        # Simple case: two ring graphs
        g1 = Graph(4, [(0, 1), (1, 2), (2, 3), (3, 0)])
        g2 = Graph(4, [(0, 1), (1, 3), (3, 2), (2, 0)])

        cp = g1.canonical_permutation()
        g3 = g1.permute_vertices(cp)

        cp = g2.canonical_permutation()
        g4 = g2.permute_vertices(cp)

        self.assertTrue(g3.vcount() == g4.vcount())
        self.assertTrue(sorted(g3.get_edgelist()) == sorted(g4.get_edgelist()))

        # Simple case with coloring
        cp = g1.canonical_permutation(color=[0, 0, 1, 1])
        g3 = g1.permute_vertices(cp)

        cp = g2.canonical_permutation(color=[0, 0, 1, 1])
        g4 = g2.permute_vertices(cp)

        self.assertTrue(g3.vcount() == g4.vcount())
        self.assertTrue(sorted(g3.get_edgelist()) == sorted(g4.get_edgelist()))

        # More complicated one: small GRG, random permutation
        g = Graph.GRG(10, 0.5)
        perm = list(range(10))
        shuffle(perm)
        g2 = g.permute_vertices(perm)
        g3 = g.permute_vertices(g.canonical_permutation())
        g4 = g2.permute_vertices(g2.canonical_permutation())

        self.assertTrue(g3.vcount() == g4.vcount())
        self.assertTrue(sorted(g3.get_edgelist()) == sorted(g4.get_edgelist()))

    def testPermuteVertices(self):
        g1 = Graph(
            8,
            [
                (0, 4),
                (0, 5),
                (0, 6),
                (1, 4),
                (1, 5),
                (1, 7),
                (2, 4),
                (2, 6),
                (2, 7),
                (3, 5),
                (3, 6),
                (3, 7),
            ],
        )
        g2 = Graph(
            8,
            [
                (0, 1),
                (0, 3),
                (0, 4),
                (2, 3),
                (2, 1),
                (2, 6),
                (5, 1),
                (5, 4),
                (5, 6),
                (7, 3),
                (7, 6),
                (7, 4),
            ],
        )
        _, _, mapping = g1.isomorphic_vf2(g2, return_mapping_21=True)
        g3 = g2.permute_vertices(mapping)
        self.assertTrue(g3.vcount() == g2.vcount() and g3.ecount() == g2.ecount())
        self.assertTrue(set(g3.get_edgelist()) == set(g1.get_edgelist()))


class AutomorphismTests(unittest.TestCase):
    def testCountAutomorphisms(self):
        g = Graph.Famous("petersen")
        self.assertEqual(120, g.count_automorphisms())

        g = Graph.Lattice([16, 16])
        self.assertEqual(2048, g.count_automorphisms())

    def testAutomorphismGroup(self):
        g = Graph.Famous("petersen")
        generators = g.automorphism_group()
        generators.sort()
        self.assertEqual(
            generators,
            [
                [0, 1, 2, 7, 5, 4, 6, 3, 9, 8],
                [0, 4, 3, 8, 5, 1, 9, 2, 6, 7],
                [1, 2, 3, 8, 6, 0, 7, 4, 5, 9],
            ],
        )


def suite():
    isomorphism_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        IsomorphismTests
    )
    subisomorphism_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        SubisomorphismTests
    )
    permutation_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        PermutationTests
    )
    automorphism_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        AutomorphismTests
    )
    return unittest.TestSuite(
        [
            isomorphism_suite,
            subisomorphism_suite,
            permutation_suite,
            automorphism_suite,
        ]
    )


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



def _construct_bipartite_graph_from_adjacency(
    cls,
    matrix,
    directed=False,
    mode="out",
    multiple=False,
    weighted=None,
    *args,
    **kwds
):
    """Creates a bipartite graph from a bipartite adjacency matrix.

    Example:

    >>> g = Graph.Biadjacency([[0, 1, 1], [1, 1, 0]])

    @param matrix: the bipartite adjacency matrix.
    @param directed: whether to create a directed graph.
    @param mode: defines the direction of edges in the graph. If
      C{"out"}, then edges go from vertices of the first kind
      (corresponding to rows of the matrix) to vertices of the
      second kind (the columns of the matrix). If C{"in"}, the
      opposite direction is used. C{"all"} creates mutual edges.
      Ignored for undirected graphs.
    @param multiple: defines what to do with non-zero entries in the
      matrix. If C{False}, non-zero entries will create an edge no matter
      what the value is. If C{True}, non-zero entries are rounded up to
      the nearest integer and this will be the number of multiple edges
      created.
    @param weighted: defines whether to create a weighted graph from the
      adjacency matrix. If it is c{None} then an unweighted graph is created
      and the multiple argument is used to determine the edges of the graph.
      If it is a string then for every non-zero matrix entry, an edge is created
      and the value of the entry is added as an edge attribute named by the
      weighted argument. If it is C{True} then a weighted graph is created and
      the name of the edge attribute will be C{"weight"}.

    @raise ValueError: if the weighted and multiple are passed together.

    @return: the graph with a binary vertex attribute named C{"type"} that
      stores the vertex classes.
    """
    is_weighted = True if weighted or weighted == "" else False
    if is_weighted and multiple:
        raise ValueError("arguments weighted and multiple can not co-exist")
    result, types = cls._Biadjacency(matrix, directed, mode, multiple, *args, **kwds)
    result.vs["type"] = types
    if is_weighted:
        weight_attr = "weight" if weighted is True else weighted
        _, rows, _ = result.get_biadjacency()
        num_vertices_of_first_kind = len(rows)
        for edge in result.es:
            source, target = edge.tuple
            if source in rows:
                edge[weight_attr] = matrix[source][target - num_vertices_of_first_kind]
            else:
                edge[weight_attr] = matrix[target][source - num_vertices_of_first_kind]
    return result


def _construct_bipartite_graph(cls, types, edges, directed=False, *args, **kwds):
    """Creates a bipartite graph with the given vertex types and edges.
    This is similar to the default constructor of the graph, the
    only difference is that it checks whether all the edges go
    between the two vertex classes and it assigns the type vector
    to a C{type} attribute afterwards.

    Examples:

    >>> g = Graph.Bipartite([0, 1, 0, 1], [(0, 1), (2, 3), (0, 3)])
    >>> g.is_bipartite()
    True
    >>> g.vs["type"]
    [False, True, False, True]

    @param types: the vertex types as a boolean list. Anything that
      evaluates to C{False} will denote a vertex of the first kind,
      anything that evaluates to C{True} will denote a vertex of the
      second kind.
    @param edges: the edges as a list of tuples.
    @param directed: whether to create a directed graph. Bipartite
      networks are usually undirected, so the default is C{False}

    @return: the graph with a binary vertex attribute named C{"type"} that
      stores the vertex classes.
    """
    result = cls._Bipartite(types, edges, directed, *args, **kwds)
    result.vs["type"] = [bool(x) for x in types]
    return result


def _construct_full_bipartite_graph(
    cls, n1, n2, directed=False, mode="all", *args, **kwds
):
    """Generates a full bipartite graph (directed or undirected, with or
    without loops).

    >>> g = Graph.Full_Bipartite(2, 3)
    >>> g.is_bipartite()
    True
    >>> g.vs["type"]
    [False, False, True, True, True]

    @param n1: the number of vertices of the first kind.
    @param n2: the number of vertices of the second kind.
    @param directed: whether tp generate a directed graph.
    @param mode: if C{"out"}, then all vertices of the first kind are
      connected to the others; C{"in"} specifies the opposite direction,
      C{"all"} creates mutual edges. Ignored for undirected graphs.

    @return: the graph with a binary vertex attribute named C{"type"} that
      stores the vertex classes.
    """
    result, types = cls._Full_Bipartite(n1, n2, directed, mode, *args, **kwds)
    result.vs["type"] = types
    return result


def _construct_random_bipartite_graph(
    cls, n1, n2, p=None, m=None, directed=False, neimode="all", *args, **kwds
):
    """Generates a random bipartite graph with the given number of vertices and
    edges (if m is given), or with the given number of vertices and the given
    connection probability (if p is given).

    If m is given but p is not, the generated graph will have n1 vertices of
    type 1, n2 vertices of type 2 and m randomly selected edges between them. If
    p is given but m is not, the generated graph will have n1 vertices of type 1
    and n2 vertices of type 2, and each edge will exist between them with
    probability p.

    @param n1: the number of vertices of type 1.
    @param n2: the number of vertices of type 2.
    @param p: the probability of edges. If given, C{m} must be missing.
    @param m: the number of edges. If given, C{p} must be missing.
    @param directed: whether to generate a directed graph.
    @param neimode: if the graph is directed, specifies how the edges will be
      generated. If it is C{"all"}, edges will be generated in both directions
      (from type 1 to type 2 and vice versa) independently. If it is C{"out"}
      edges will always point from type 1 to type 2. If it is C{"in"}, edges
      will always point from type 2 to type 1. This argument is ignored for
      undirected graphs.
    """
    if p is None:
        p = -1
    if m is None:
        m = -1
    result, types = cls._Random_Bipartite(
        n1, n2, p, m, directed, neimode, *args, **kwds
    )
    result.vs["type"] = types
    return result



from typing import Union, Sequence
from collections import defaultdict
from itertools import repeat
from warnings import warn

from igraph.datatypes import UniqueIdGenerator


def _construct_graph_from_dict_list(
    cls,
    vertices,
    edges,
    directed: bool = False,
    vertex_name_attr: str = "name",
    edge_foreign_keys=("source", "target"),
    iterative: bool = False,
):
    """Constructs a graph from a list-of-dictionaries representation.

    This function is useful when you have two lists of dictionaries, one for
    vertices and one for edges, each containing their attributes (e.g. name,
    weight). Of course, the edge dictionary must also contain two special keys
    that indicate the source and target vertices connected by that edge.
    Non-list iterables should work as long as they yield dictionaries or
    dict-like objects (they should have the 'items' and '__getitem__' methods).
    For instance, a database query result is likely to be fit as long as it's
    iterable and yields dict-like objects with every iteration.

    @param vertices: the list of dictionaries for the vertices or C{None} if
      there are no special attributes assigned to vertices and we
      should simply use the edge list of dicts to infer vertex names.
    @param edges: the list of dictionaries for the edges. Each dict must have
      at least the two keys specified by edge_foreign_keys to label the source
      and target vertices, while additional items will be treated as edge
      attributes.
    @param directed: whether the constructed graph will be directed
    @param vertex_name_attr: the name of the distinguished key in the
      dicts in the vertex data source that contains the vertex names.
      Ignored if C{vertices} is C{None}.
    @param edge_foreign_keys: tuple specifying the attributes in each edge
      dictionary that contain the source (1st) and target (2nd) vertex names.
      These items of each dictionary are also added as edge_attributes.
    @param iterative: whether to add the edges to the graph one by one,
      iteratively, or to build a large edge list first and use that to
      construct the graph. The latter approach is faster but it may
      not be suitable if your dataset is large. The default is to
      add the edges in a batch from an edge list.
    @return: the graph that was constructed

    Example:

    >>> vertices = [{'name': 'apple'}, {'name': 'pear'}, {'name': 'peach'}]
    >>> edges = [{'source': 'apple', 'target': 'pear', 'weight': 1.2},
    ...          {'source': 'apple', 'target': 'peach', 'weight': 0.9}]
    >>> g = Graph.DictList(vertices, edges)

    The graph has three vertices with names and two edges with weights.
    """

    def create_list_from_indices(indices, n):
        result = [None] * n
        for i, v in indices:
            result[i] = v
        return result

    # Construct the vertices
    vertex_attrs = {}
    n = 0
    if vertices:
        for idx, vertex_data in enumerate(vertices):
            for k, v in vertex_data.items():
                try:
                    vertex_attrs[k].append((idx, v))
                except KeyError:
                    vertex_attrs[k] = [(idx, v)]
            n += 1
        for k, v in vertex_attrs.items():
            vertex_attrs[k] = create_list_from_indices(v, n)
    else:
        vertex_attrs[vertex_name_attr] = []

    if vertex_name_attr not in vertex_attrs:
        raise AttributeError(
            f"{vertex_name_attr} is not a key of your vertex dictionaries",
        )
    vertex_names = vertex_attrs[vertex_name_attr]

    # Check for duplicates in vertex_names
    if len(vertex_names) != len(set(vertex_names)):
        raise ValueError("vertex names are not unique")
    # Create a reverse mapping from vertex names to indices
    vertex_name_map = UniqueIdGenerator(initial=vertex_names)

    # Construct the edges
    efk_src, efk_dest = edge_foreign_keys
    if iterative:
        g = cls(n, [], directed, {}, vertex_attrs)
        for idx, edge_data in enumerate(edges):
            src_name = edge_data[efk_src]
            dst_name = edge_data[efk_dest]
            v1 = vertex_name_map[src_name]
            if v1 == n:
                g.add_vertices(1)
                g.vs[n][vertex_name_attr] = src_name
                n += 1
            v2 = vertex_name_map[dst_name]
            if v2 == n:
                g.add_vertices(1)
                g.vs[n][vertex_name_attr] = dst_name
                n += 1
            g.add_edge(v1, v2)
            for k, v in edge_data.items():
                g.es[idx][k] = v

        return g
    else:
        edge_list = []
        edge_attrs = {}
        m = 0
        for idx, edge_data in enumerate(edges):
            v1 = vertex_name_map[edge_data[efk_src]]
            v2 = vertex_name_map[edge_data[efk_dest]]

            edge_list.append((v1, v2))
            for k, v in edge_data.items():
                try:
                    edge_attrs[k].append((idx, v))
                except KeyError:
                    edge_attrs[k] = [(idx, v)]
            m += 1
        for k, v in edge_attrs.items():
            edge_attrs[k] = create_list_from_indices(v, m)

        # It may have happened that some vertices were added during
        # the process
        if len(vertex_name_map) > n:
            diff = len(vertex_name_map) - n
            more = [None] * diff
            for v in vertex_attrs.values():
                v.extend(more)
            vertex_attrs[vertex_name_attr] = list(vertex_name_map.values())
            n = len(vertex_name_map)

        # Create the graph
        return cls(n, edge_list, directed, {}, vertex_attrs, edge_attrs)


def _construct_graph_from_tuple_list(
    cls,
    edges,
    directed: bool = False,
    vertex_name_attr: str = "name",
    edge_attrs=None,
    weights=False,
):
    """Constructs a graph from a list-of-tuples representation.

    This representation assumes that the edges of the graph are encoded
    in a list of tuples (or lists). Each item in the list must have at least
    two elements, which specify the source and the target vertices of the edge.
    The remaining elements (if any) specify the edge attributes of that edge,
    where the names of the edge attributes originate from the C{edge_attrs}
    list. The names of the vertices will be stored in the vertex attribute
    given by C{vertex_name_attr}.

    The default parameters of this function are suitable for creating
    unweighted graphs from lists where each item contains the source vertex
    and the target vertex. If you have a weighted graph, you can use items
    where the third item contains the weight of the edge by setting
    C{edge_attrs} to C{"weight"} or C{["weight"]}. If you have even more
    edge attributes, add them to the end of each item in the C{edges}
    list and also specify the corresponding edge attribute names in
    C{edge_attrs} as a list.

    @param edges: the data source for the edges. This must be a list
      where each item is a tuple (or list) containing at least two
      items: the name of the source and the target vertex. Note that
      names will be assigned to the C{name} vertex attribute (or another
      vertex attribute if C{vertex_name_attr} is specified), even if
      all the vertex names in the list are in fact numbers.
    @param directed: whether the constructed graph will be directed
    @param vertex_name_attr: the name of the vertex attribute that will
      contain the vertex names.
    @param edge_attrs: the names of the edge attributes that are filled
      with the extra items in the edge list (starting from index 2, since
      the first two items are the source and target vertices). If C{None}
      or an empty sequence, only the source and target vertices will be
      extracted and additional tuple items will be ignored. If a string, it is
      interpreted as a single edge attribute.
    @param weights: alternative way to specify that the graph is
      weighted. If you set C{weights} to C{true} and C{edge_attrs} is
      not given, it will be assumed that C{edge_attrs} is C{["weight"]}
      and igraph will parse the third element from each item into an
      edge weight. If you set C{weights} to a string, it will be assumed
      that C{edge_attrs} contains that string only, and igraph will
      store the edge weights in that attribute.
    @return: the graph that was constructed
    """
    if edge_attrs is None:
        if not weights:
            edge_attrs = ()
        else:
            if not isinstance(weights, str):
                weights = "weight"
            edge_attrs = [weights]
    else:
        if weights:
            raise ValueError("`weights` must be False if `edge_attrs` is " "not None")

    if isinstance(edge_attrs, str):
        edge_attrs = [edge_attrs]

    # Set up a vertex ID generator
    idgen = UniqueIdGenerator()

    # Construct the edges and the edge attributes
    edge_list = []
    edge_attributes = {}
    for name in edge_attrs:
        edge_attributes[name] = []

    for item in edges:
        edge_list.append((idgen[item[0]], idgen[item[1]]))
        for index, name in enumerate(edge_attrs, 2):
            try:
                edge_attributes[name].append(item[index])
            except IndexError:
                edge_attributes[name].append(None)

    # Set up the name vertex attribute
    vertex_attributes = {}
    vertex_attributes[vertex_name_attr] = list(idgen.values())
    n = len(idgen)

    # Construct the graph
    return cls(n, edge_list, directed, {}, vertex_attributes, edge_attributes)


def _construct_graph_from_list_dict(
    cls,
    edges,
    directed: bool = False,
    vertex_name_attr: str = "name",
):
    """Constructs a graph from a dict-of-lists representation.

    This function is used to construct a graph from a dictionary of
    lists. Other, non-list sequences (e.g. tuples) and lazy iterators are
    are accepted. For each key x, its corresponding value must be a sequence of
    multiple values y: the edge (x,y) will be created in the graph. x and y
    must be either one of:

      - two integers: the vertices with those ids will be connected
      - two strings: the vertices with those names will be connected

    If names are used, the order of vertices is not guaranteed, and each
    vertex will be given the vertex_name_attr attribute.

    @param edges: the dict of sequences describing the edges
    @param directed: whether to create a directed graph
    @param vertex_name_attr: vertex attribute that will store the names

    @returns: a Graph object

    Example:

    >>> mydict = {'apple': ['pear', 'peach'], 'pear': ['peach']}
    >>> g = Graph.ListDict(mydict)

    # The graph has three vertices with names and three edges connecting
    # each pair.
    """
    first_item = next(iter(edges), 0)

    if not isinstance(first_item, (int, str)):
        raise ValueError("Keys must be integers or strings")

    vertex_attributes = {}
    if isinstance(first_item, str):
        name_map = UniqueIdGenerator()
        edge_list = []
        for source, sequence in edges.items():
            source_id = name_map[source]
            edge_list.extend((source_id, name_map[target]) for target in sequence)
        vertex_attributes[vertex_name_attr] = name_map.values()
        n = len(name_map)

    else:
        edge_list = []
        n = -1
        for source, sequence in edges.items():
            n = max(n, source, *sequence)
            edge_list.extend(zip(repeat(source), sequence))
        n += 1

    # Construct the graph
    return cls(n, edge_list, directed, {}, vertex_attributes, {})


def _construct_graph_from_dict_dict(
    cls,
    edges,
    directed: bool = False,
    vertex_name_attr: str = "name",
):
    """Constructs a graph from a dict-of-dicts representation.

    Each key can be an integer or a string and represent a vertex. Each value
    is a dict representing edges (outgoing if the graph is directed) from that
    vertex. Each dict key is an integer/string for a target vertex, such that
    an edge will be created between those two vertices. Integers are
    interpreted as vertex_ids from 0 (as used in igraph), strings are
    interpreted as vertex names, in which case vertices are given separate
    numeric ids. Each value is a dictionary of edge attributes for that edge.

    Example:

      >>> {'Alice': {'Bob': {'weight': 1.5}, 'David': {'weight': 2}}}

    creates a graph with three vertices (Alice, Bob, and David) and two edges:

      - Alice - Bob (with weight 1.5)
      - Alice - David (with weight 2)

    @param edges: the dict of dict of dicts specifying the edges and their
      attributes
    @param directed: whether to create a directed graph
    @param vertex_name_attr: vertex attribute that will store the names

    @returns: a Graph object
    """
    first_item = next(iter(edges), 0)

    if not isinstance(first_item, (int, str)):
        raise ValueError("Keys must be integers or strings")

    vertex_attributes = {}
    edge_attribute_list = []
    if isinstance(first_item, str):
        name_map = UniqueIdGenerator()
        edge_list = []
        for source, target_dict in edges.items():
            source_id = name_map[source]
            for target, edge_attrs in target_dict.items():
                edge_list.append((source_id, name_map[target]))
                edge_attribute_list.append(edge_attrs)
        vertex_attributes[vertex_name_attr] = name_map.values()
        n = len(name_map)

    else:
        edge_list = []
        n = -1
        for source, target_dict in edges.items():
            n = max(n, source, *target_dict)
            for target, edge_attrs in target_dict.items():
                edge_list.append((source, target))
                edge_attribute_list.append(edge_attrs)
        n += 1

    # Construct graph without edge attributes
    graph = cls(n, edge_list, directed, {}, vertex_attributes, {})

    # Add edge attributes
    for edge, edge_attrs in zip(graph.es, edge_attribute_list):
        for key, val in edge_attrs.items():
            edge[key] = val

    return graph


def _construct_graph_from_dataframe(
    cls,
    edges,
    directed: bool = True,
    vertices=None,
    use_vids: bool = True,
):
    """Generates a graph from one or two dataframes.

    @param edges: pandas DataFrame containing edges and metadata. The first
      two columns of this DataFrame contain the source and target vertices
      for each edge. These indicate the vertex IDs as nonnegative integers
      rather than vertex names unless C{use_vids} is False. Further columns
      may contain edge attributes.
    @param directed: whether the graph is directed
    @param vertices: None (default) or pandas DataFrame containing vertex
      metadata. The DataFrame's index must contain the vertex IDs as a
      sequence of intergers from 0 to C{len(vertices) - 1}. If C{use_vids}
      is C{False}, the first column must contain the unique vertex names.
      Vertex names should be strings for full compatibility, but many functions
      will work if you set the name with any hashable object. All other columns
      will be added as vertex attributes by column name.
    @param use_vids: whether to interpret the first two columns of the C{edges}
      argument as vertex ids (0-based integers) instead of vertex names.
      If this argument is set to True and the first two columns of C{edges}
      are not integers, an error is thrown.

    @return: the graph

    Vertex names in either the C{edges} or C{vertices} arguments that are set
    to NaN (not a number) will be set to the string "NA". That might lead
    to unexpected behaviour: fill your NaNs with values before calling this
    function to mitigate.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "You should install pandas in order to use this function"
        ) from None

    if edges.shape[1] < 2:
        raise ValueError("The 'edges' DataFrame must contain at least two columns")
    if vertices is not None and vertices.shape[1] < 1:
        raise ValueError("The 'vertices' DataFrame must contain at least one column")

    if use_vids:
        if str(edges.dtypes.iloc[0]).startswith(("int", "Int")) and str(
            edges.dtypes.iloc[1]
        ).startswith(("int", "Int")):
            # Check pandas nullable integer data type:
            # https://pandas.pydata.org/docs/user_guide/integer_na.html
            if (edges.iloc[:, :2].isna()).any(axis=None):
                raise ValueError("Source and target IDs must not be null")

            if (edges.iloc[:, :2] < 0).any(axis=None):
                raise ValueError("Source and target IDs must not be negative")
        else:
            raise TypeError(
                f"Source and target IDs must be 0-based integers, found types {edges.dtypes.tolist()[:2]}"
            )

        if vertices is not None:
            vertices = vertices.sort_index()
            if not vertices.index.equals(
                pd.RangeIndex.from_range(range(vertices.shape[0]))
            ):
                if not str(vertices.index.dtype).startswith("int"):
                    raise TypeError(
                        f"Vertex IDs must be 0-based integers, found type {vertices.index.dtype}"
                    )
                elif (vertices.index < 0).any(axis=None):
                    raise ValueError("Vertex IDs must not be negative")
                else:
                    raise ValueError(
                        f"Vertex IDs must be an integer sequence from 0 to {vertices.shape[0] - 1}"
                    )
    else:
        # Handle if some source and target names in 'edges' are 'NA'
        if edges.iloc[:, :2].isna().any(axis=None):
            warn(
                "In the first two columns of 'edges' NA elements were replaced with string \"NA\"",
                stacklevel=1,
            )
            edges = edges.copy()
            edges.iloc[:, :2].fillna("NA", inplace=True)

        # Bring DataFrame(s) into same format as with 'use_vids=True'
        if vertices is None:
            vertices = pd.DataFrame({"name": pd.unique(edges.values[:, :2].ravel())})

        if vertices.iloc[:, 0].isna().any():
            warn(
                "In the first column of 'vertices' NA elements were replaced with string \"NA\"",
                stacklevel=1,
            )
            vertices = vertices.copy()
            vertices.iloc[:, 0].fillna("NA", inplace=True)

        if vertices.iloc[:, 0].duplicated().any():
            raise ValueError("Vertex names must be unique")

        if vertices.shape[1] > 1 and "name" in vertices.columns[1:]:
            raise ValueError(
                "Vertex attribute conflict: DataFrame already contains column 'name'"
            )

        vertices = vertices.rename({vertices.columns[0]: "name"}, axis=1).reset_index(
            drop=True
        )

        # Map source and target names in 'edges' to IDs
        vid_map = pd.Series(vertices.index, index=vertices.iloc[:, 0])
        edges = edges.copy()
        edges[edges.columns[0]] = edges.iloc[:, 0].map(vid_map)
        edges[edges.columns[1]] = edges.iloc[:, 1].map(vid_map)

    # Create graph
    if vertices is None:
        nv = edges.iloc[:, :2].max().max() + 1
        g = cls(n=nv, directed=directed)
    else:
        if not edges.iloc[:, :2].isin(vertices.index).all(axis=None):
            raise ValueError(
                "Some vertices in the edge DataFrame are missing from vertices DataFrame"
            )
        nv = vertices.shape[0]
        g = cls(n=nv, directed=directed)
        # Add vertex attributes
        for col in vertices.columns:
            g.vs[col] = vertices[col].tolist()

    # add edges including optional attributes
    e_list = list(edges.iloc[:, :2].itertuples(index=False, name=None))
    e_attr = edges.iloc[:, 2:].to_dict(orient="list") if edges.shape[1] > 2 else None
    g.add_edges(e_list, e_attr)

    return g


def _export_graph_to_dict_list(
    graph,
    use_vids: bool = True,
    skip_none: bool = False,
    vertex_name_attr: str = "name",
):
    """Export graph as two lists of dictionaries, for vertices and edges.

    This function is the reverse of Graph.DictList.

    Example:

      >>> g = Graph([(0, 1), (1, 2)])
      >>> g.vs["name"] = ["apple", "pear", "peach"]
      >>> g.es["name"] = ["first_edge", "second"]

      >>> g.to_dict_list()
      ([{"name": "apple"}, {"name": "pear"}, {"name": "peach"}],
       [{"source": 0, "target": 1, "name": "first_edge"},
        {"source" 0, "target": 2, name": "second"}])

      >>> g.to_dict_list(use_vids=False)
      ([{"name": "apple"}, {"name": "pear"}, {"name": "peach"}],
       [{"source": "apple", "target": "pear", "name": "first_edge"},
        {"source" "apple", "target": "peach", name": "second"}])

    @param use_vids: whether to label vertices in the output data
      structure by their ids or their vertex_name_attr attribute. If
      use_vids=False but vertices lack a vertex_name_attr attribute, an
      AttributeError is raised.
    @param skip_none: whether to skip, for each edge, attributes that
      have a value of None. This is useful if only some edges are expected to
      possess an attribute.
    @param vertex_name_attr: only used with use_vids=False to choose what
      vertex attribute to use to name your vertices in the output data
      structure.

    @return: a tuple with two lists of dictionaries, representing the vertices
      and the edges, respectively, with their attributes.
    """
    # Output data structures
    res_vs, res_es = [], []

    if not use_vids:
        if vertex_name_attr not in graph.vertex_attributes():
            raise AttributeError(f"No vertex attribute {vertex_name_attr}")

        vs_names = graph.vs[vertex_name_attr]

    for vertex in graph.vs:
        if skip_none:
            attrdic = {k: v for k, v in vertex.attributes() if v is not None}
        else:
            attrdic = vertex.attributes()
        res_vs.append(attrdic)

    for edge in graph.es:
        source, target = edge.tuple
        if not use_vids:
            source, target = vs_names[source], vs_names[target]
        if skip_none:
            attrdic = {k: v for k, v in edge.attributes() if v is not None}
        else:
            attrdic = edge.attributes()

        attrdic["source"] = source
        attrdic["target"] = target
        res_es.append(attrdic)

    return (res_vs, res_es)


def _export_graph_to_tuple_list(
    graph,
    use_vids: bool = True,
    edge_attrs: Union[str, Sequence[str]] = None,
    vertex_name_attr: str = "name",
):
    """Export graph to a list of edge tuples

    This function is the reverse of Graph.TupleList.

    Example:

      >>> g = Graph.Full(3)
      >>> g.vs["name"] = ["apple", "pear", "peach"]
      >>> g.es["name"] = ["first_edge", "second", "third"]

      >>> # Get name of the edge
      >>> g.to_tuple_list(edge_attrs=["name"])
      [(0, 1, "first_edge"), (0, 2, "second"), (1, 2, "third")]

      >>> # Use vertex names, no edge attributes
      >>> g.to_tuple_list(use_vids=False)
      [("apple", "pear"), ("apple", "peach"), ("pear", "peach")]

    @param use_vids: whether to label vertices in the output data
      structure by their ids or their vertex_name_attr attribute. If
      use_vids=False but vertices lack a vertex_name_attr attribute, an
      AttributeError is raised.
    @param edge_attrs: list of edge attributes to export
      in addition to source and target vertex, which are always the first two
      elements of each tuple. None (default) is equivalent to an empty list. A
      string is acceptable to signify a single attribute and will be wrapped in
      a list internally.
    @param vertex_name_attr: only used with use_vids=False to choose what
      vertex attribute to use to name your vertices in the output data
      structure.

    @return: a list of tuples, each representing an edge of the graph.
    """
    # Output data structure
    res = []

    if edge_attrs is not None:
        if isinstance(edge_attrs, str):
            edge_attrs = [edge_attrs]
        missing_attrs = list(set(edge_attrs) - set(graph.edge_attributes()))
        if missing_attrs:
            raise AttributeError(f"Missing attributes: {missing_attrs}")
    else:
        edge_attrs = []

    if use_vids is False:
        if vertex_name_attr not in graph.vertex_attributes():
            raise AttributeError(f"No vertex attribute {vertex_name_attr}")

        vs_names = graph.vs[vertex_name_attr]

    for edge in graph.es:
        source, target = edge.tuple
        if not use_vids:
            source, target = vs_names[source], vs_names[target]
        attrlist = [source, target]
        attrlist += [edge[attrname] for attrname in edge_attrs]
        res.append(tuple(attrlist))

    return res


def _export_graph_to_list_dict(
    graph,
    use_vids: bool = True,
    sequence_constructor: callable = list,
    vertex_name_attr: str = "name",
):
    """Export graph to a dictionary of lists (or other sequences).

    This function is the reverse of Graph.ListDict.

    Example:

      >>> g = Graph.Full(3)
      >>> g.to_sequence_dict() -> {0: [1, 2], 1: [2]}
      >>> g.to_sequence_dict(sequence_constructor=tuple) -> {0: (1, 2), 1: (2,)}
      >>> g.vs['name'] = ['apple', 'pear', 'peach']
      >>> g.to_sequence_dict(use_vids=False)
      {'apple': ['pear', 'peach'], 'pear': ['peach']}

    @param use_vids: whether to label vertices in the output data
      structure by their ids or their vertex_name_attr attribute. If
      use_vids=False but vertices lack a vertex_name_attr attribute, an
      AttributeError is raised.
    @param sequence_constructor: constructor for the data structure
      to be used as values of the dictionary. The default (list) makes a dict
      of lists, with each list representing the neighbors of the vertex
      specified in the respective dictionary key.
    @param vertex_name_attr: only used with use_vids=False to choose what
      vertex attribute to use to name your vertices in the output data
      structure.

    @return: dictionary of sequences, keyed by vertices, with each value
      containing the neighbors of that vertex.
    """
    if not use_vids:
        if vertex_name_attr not in graph.vertex_attributes():
            raise AttributeError(f"Vertices do not have a {vertex_name_attr} attribute")
        vs_names = graph.vs[vertex_name_attr]

    # Temporary output data structure
    res = defaultdict(list)

    for edge in graph.es:
        source, target = edge.tuple

        if not use_vids:
            source = vs_names[source]
            target = vs_names[target]

        res[source].append(target)

    res = {key: sequence_constructor(val) for key, val in res.items()}
    return res


def _export_graph_to_dict_dict(
    graph,
    use_vids: bool = True,
    edge_attrs: Union[str, Sequence[str]] = None,
    skip_none: bool = False,
    vertex_name_attr: str = "name",
):
    """Export graph to dictionary of dicts of edge attributes

    This function is the reverse of Graph.DictDict.

    Example:

      >>> g = Graph.Full(3)
      >>> g.es['name'] = ['first_edge', 'second', 'third']
      >>> g.to_dict_dict()
      {0: {1: {'name': 'first_edge'}, 2: {'name': 'second'}}, 1: {2: {'name': 'third'}}}

    @param use_vids: whether to label vertices in the output data
      structure by their ids or their vertex_name_attr attribute. If
      use_vids=False but vertices lack a vertex_name_attr attribute, an
      AttributeError is raised.
    @param edge_attrs: list of edge attributes to export.
      None (default) signified all attributes (unlike Graph.to_tuple_list). A
      string is acceptable to signify a single attribute and will be wrapped
      in a list internally.
    @param skip_none: whether to skip, for each edge, attributes that
      have a value of None. This is useful if only some edges are expected to
      possess an attribute.
    @param vertex_name_attr: only used with use_vids=False to choose what
      vertex attribute to use to name your vertices in the output data
      structure.

    @return: dictionary of dictionaries of dictionaries, with the outer keys
      vertex ids/names, the middle keys ids/names of their neighbors, and the
      innermost dictionary representing attributes of that edge.
    """
    if edge_attrs is not None:
        if isinstance(edge_attrs, str):
            edge_attrs = [edge_attrs]
        missing_attrs = list(set(edge_attrs) - set(graph.edge_attributes()))
        if missing_attrs:
            raise AttributeError(f"Missing attributes: {missing_attrs}")

    if not use_vids:
        if vertex_name_attr not in graph.vertex_attributes():
            raise AttributeError(f"Vertices do not have a {vertex_name_attr} attribute")
        vs_names = graph.vs[vertex_name_attr]

    # Temporary output data structure
    res = defaultdict(lambda: defaultdict(dict))

    for edge in graph.es:
        source, target = edge.tuple

        if not use_vids:
            source = vs_names[source]
            target = vs_names[target]

        attrdic = edge.attributes()
        if edge_attrs is not None:
            attrdic = {k: attrdic[k] for k in edge_attrs}
        if skip_none:
            attrdic = {k: v for k, v in attrdic.items() if v is not None}

        res[source][target] = attrdic

    res = {key: dict(val) for key, val in res.items()}
    return res


def _export_vertex_dataframe(graph):
    """Export vertices with attributes to pandas.DataFrame

    If you want to use vertex names as index, you can do:

    >>> from string import ascii_letters
    >>> graph = Graph.GRG(25, 0.4)
    >>> graph.vs["name"] = ascii_letters[:graph.vcount()]
    >>> df = graph.get_vertex_dataframe()
    >>> df.set_index('name', inplace=True)

    @return: a pandas.DataFrame representing vertices and their attributes.
      The index uses vertex IDs, from 0 to N - 1 where N is the number of
      vertices.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "You should install pandas in order to use this function"
        ) from None

    df = pd.DataFrame(
        {attr: graph.vs[attr] for attr in graph.vertex_attributes()},
        index=list(range(graph.vcount())),
    )
    df.index.name = "vertex ID"

    return df


def _export_edge_dataframe(graph):
    """Export edges with attributes to pandas.DataFrame

    If you want to use source and target vertex IDs as index, you can do:

    >>> from string import ascii_letters
    >>> graph = Graph.GRG(25, 0.4)
    >>> graph.vs["name"] = ascii_letters[:graph.vcount()]
    >>> df = graph.get_edge_dataframe()
    >>> df.set_index(['source', 'target'], inplace=True)

    The index will be a pandas.MultiIndex. You can use the C{drop=False}
    option to keep the C{source} and C{target} columns.

    If you want to use vertex names in the source and target columns:

    >>> df = graph.get_edge_dataframe()
    >>> df_vert = graph.get_vertex_dataframe()
    >>> df['source'].replace(df_vert['name'], inplace=True)
    >>> df['target'].replace(df_vert['name'], inplace=True)
    >>> df_vert.set_index('name', inplace=True)  # Optional

    @return: a pandas.DataFrame representing edges and their attributes.
      The index uses edge IDs, from 0 to M - 1 where M is the number of
      edges. The first two columns of the dataframe represent the IDs of
      source and target vertices for each edge. These columns have names
      "source" and "target". If your edges have attributes with the same
      names, they will be present in the dataframe, but not in the first
      two columns.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "You should install pandas in order to use this function"
        ) from None

    df = pd.DataFrame(
        {attr: graph.es[attr] for attr in graph.edge_attributes()},
        index=list(range(graph.ecount())),
    )
    df.index.name = "edge ID"

    df.insert(0, "source", [e.source for e in graph.es], allow_duplicates=True)
    df.insert(1, "target", [e.target for e in graph.es], allow_duplicates=True)

    return df



"""
Drawers for labels on plots.
"""

from enum import Enum

__all__ = ("TextAlignment",)

#####################################################################


class TextAlignment(Enum):
    """Text alignment constants."""

    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"
    TOP = "top"
    BOTTOM = "bottom"



"""This module provides implementation for a Matplotlib-specific palette drawer."""

from igraph.drawing.baseclasses import AbstractDrawer

__all__ = ("MatplotlibPaletteDrawer",)


class MatplotlibPaletteDrawer(AbstractDrawer):
    """Matplotlib drawer object for matrices."""

    def __init__(self, ax):
        """Constructs the drawer and associates it to the given Axes.

        @param ax: the Axes on which we will draw
        """
        self.context = ax

    def draw(self, matrix, **kwds):
        """Draws the given Matrix in a matplotlib Axes.

        @param matrix: the igraph.Histogram to plot.

        """
        from igraph.datatypes import Matrix
        from igraph.drawing.utils import find_matplotlib, str_to_orientation

        mpl, _ = find_matplotlib()
        ax = self.context

        orientation = str_to_orientation(kwds.get("orientation", "lr"))

        # Construct a matrix and plot that
        indices = list(range(len(self)))
        if orientation in ("rl", "bt"):
            indices.reverse()
        if orientation in ("lr", "rl"):
            matrix = Matrix([indices])
        else:
            matrix = Matrix([[i] for i in indices])

        cmap = mpl.colors.ListedColormap(
            [self.get(i) for i in range(self.length)],
        )
        matrix.__plot__(
            "matplotlib",
            ax,
            cmap=cmap,
            **kwds,
        )



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Classes that help igraph communicate with Gephi (http://www.gephi.org)."""

import json
import urllib.error
import urllib.parse
import urllib.request

__all__ = ("GephiConnection", "GephiGraphStreamer", "GephiGraphStreamingAPIFormat")
__docformat__ = "restructuredtext en"


class GephiConnection:
    """Object that represents a connection to a Gephi master server."""

    def __init__(self, url=None, host="127.0.0.1", port=8080, workspace=1):
        """Constructs a connection to a Gephi master server.

        The connection object can be constructed either by specifying the
        ``url`` directly, or by specifying the ``host``, ``port`` and
        ``workspace`` arguments.  The latter three are evaluated only if
        ``url`` is None; otherwise the ``url`` will take precedence.

        The ``url`` argument does not have to include the operation (e.g.,
        ``?operation=updateGraph``); the connection will take care of it.
        E.g., if you wish to connect to workspace 2 in a local Gephi instance
        on port 7341, the correct form to use for the ``url`` is as follows::

            http://localhost:7341/workspace0
        """
        self._pending_operations = []
        self._autoflush_threshold = 1024

        self.url = url or self._construct_default_url(host, port, workspace)

    def __del__(self):
        try:
            self.close()
        except urllib.error.URLError:
            # Happens when Gephi is closed before the connection is destroyed
            pass

    def _construct_default_url(self, host, port, workspace):
        return "http://%s:%d/workspace%d" % (host, port, workspace)

    def close(self):
        """Flushes all the pending operations to the Gephi master server in a
        single request."""
        self.flush()

    def flush(self):
        """Flushes all the pending operations to the Gephi master server in a
        single request."""
        data = b"".join(self._pending_operations)
        self._pending_operations = []
        conn = urllib.request.urlopen(self._update_url, data=data)
        return conn.read()

    @property
    def url(self):
        """The URL of the Gephi workspace where the data will be sent."""
        return self._url_root

    @url.setter
    def url(self, value):
        self._url_root = value
        self._get_url = self._url_root + "?operation=getGraph"
        self._update_url = self._url_root + "?operation=updateGraph"

    def write(self, data):
        """Sends the given raw data to the Gephi streaming master server in an HTTP
        POST request."""
        self._pending_operations.append(data)
        if len(self._pending_operations) >= self._autoflush_threshold:
            self.flush()

    def __repr__(self):
        return "%s(url=%r)" % (self.__class__.__name__, self.url)


class GephiGraphStreamingAPIFormat:
    """Object that implements the Gephi graph streaming API format and returns
    Python objects corresponding to the events defined in the API.
    """

    def get_add_node_event(self, identifier, attributes=None):
        """Generates a Python object corresponding to the event that adds a node
        with the given identifier and attributes in the Gephi graph streaming API.

        Example::

            >>> api = GephiGraphStreamingAPIFormat()
            >>> api.get_add_node_event("spam")
            {'an': {'spam': {}}}
            >>> api.get_add_node_event("spam", dict(ham="eggs"))
            {'an': {'spam': {'ham': 'eggs'}}}
        """
        return {"an": {identifier: attributes if attributes is not None else {}}}

    def get_add_edge_event(self, identifier, source, target, directed, attributes=None):
        """Generates a Python object corresponding to the event that adds an edge
        with the given source, target, directednessr and attributes in the Gephi
        graph streaming API.
        """
        result = dict(attributes if attributes is not None else {})
        result["source"] = source
        result["target"] = target
        result["directed"] = bool(directed)
        return {"ae": {identifier: result}}

    def get_change_node_event(self, identifier, attributes):
        """Generates a Python object corresponding to the event that changes the
        attributes of some node in the Gephi graph streaming API. The given attributes
        are merged into the existing ones; use C{None} as the attribute value to
        delete a given attribute.

        Example::

            >>> api = GephiGraphStreamingAPIFormat()
            >>> api.get_change_node_event("spam", dict(ham="eggs"))
            {'cn': {'spam': {'ham': 'eggs'}}}
            >>> api.get_change_node_event("spam", dict(ham=None))
            {'cn': {'spam': {'ham': None}}}
        """
        return {"cn": {identifier: attributes}}

    def get_change_edge_event(self, identifier, attributes):
        """Generates a Python object corresponding to the event that changes the
        attributes of some edge in the Gephi graph streaming API. The given attributes
        are merged into the existing ones; use C{None} as the attribute value to
        delete a given attribute.

        Example::

            >>> api = GephiGraphStreamingAPIFormat()
            >>> api.get_change_edge_event("spam", dict(ham="eggs"))
            {'ce': {'spam': {'ham': 'eggs'}}}
            >>> api.get_change_edge_event("spam", dict(ham=None))
            {'ce': {'spam': {'ham': None}}}
        """
        return {"ce": {identifier: attributes}}

    def get_delete_node_event(self, identifier):
        """Generates a Python object corresponding to the event that deletes a
        node with the given identifier in the Gephi graph streaming API.

        Example::

            >>> api = GephiGraphStreamingAPIFormat()
            >>> api.get_delete_node_event("spam")
            {'dn': {'spam': {}}}
        """
        return {"dn": {identifier: {}}}

    def get_delete_edge_event(self, identifier):
        """Generates a Python object corresponding to the event that deletes an
        edge with the given identifier in the Gephi graph streaming API.

        Example::

            >>> api = GephiGraphStreamingAPIFormat()
            >>> api.get_delete_edge_event("spam:ham")
            {'de': {'spam:ham': {}}}
        """
        return {"de": {identifier: {}}}


class GephiGraphStreamer:
    """Class that produces JSON event objects that stream an igraph graph to
    Gephi using the Gephi Graph Streaming API.

    The Gephi graph streaming format is a simple JSON-based format that can be used
    to post mutations to a graph (i.e. node and edge additions, removals and updates)
    to a remote component. For instance, one can open up Gephi (http://www.gephi.org),
    install the Gephi graph streaming plugin and then send a graph from igraph
    straight into the Gephi window by using `GephiGraphStreamer` with the
    appropriate URL where Gephi is listening.

    Example::

        >>> from cStringIO import StringIO
        >>> from igraph import Graph
        >>> buf = StringIO()
        >>> streamer = GephiGraphStreamer()
        >>> graph = Graph.Formula("A --> B, B --> C")
        >>> streamer.post(graph, buf)
        >>> print(buf.getvalue())        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        {"an": {"igraph:...:v:0": {"name": "A"}}}
        {"an": {"igraph:...:v:1": {"name": "B"}}}
        {"an": {"igraph:...:v:2": {"name": "C"}}}
        {"ae": {"igraph:...:e:0:1": {...}}}
        {"ae": {"igraph:...:e:1:2": {...}}}
        <BLANKLINE>

    """

    def __init__(self, encoder=None):
        """Constructs a Gephi graph streamer that will post graphs to a
        given file-like object or a Gephi connection.

        ``encoder`` must either be ``None`` or an instance of ``json.JSONEncoder``
        and it must contain the JSON encoder to be used when posting JSON objects.
        """
        self.encoder = encoder or json.JSONEncoder(ensure_ascii=True)
        self.format = GephiGraphStreamingAPIFormat()

    def iterjsonobj(self, graph):
        """Iterates over the JSON objects that build up the graph using the
        Gephi graph streaming API. The objects returned from this function are
        Python objects; they must be formatted with ``json.dumps`` before
        sending them to the destination.
        """

        # Construct a unique ID prefix
        id_prefix = "igraph:%s" % (hex(id(graph)),)

        # Add the vertices
        add_node = self.format.get_add_node_event
        for vertex in graph.vs:
            yield add_node("%s:v:%d" % (id_prefix, vertex.index), vertex.attributes())

        # Add the edges
        add_edge = self.format.get_add_edge_event
        directed = graph.is_directed()
        for edge in graph.es:
            yield add_edge(
                "%s:e:%d:%d" % (id_prefix, edge.source, edge.target),
                "%s:v:%d" % (id_prefix, edge.source),
                "%s:v:%d" % (id_prefix, edge.target),
                directed,
                edge.attributes(),
            )

    def post(self, graph, destination, encoder=None):
        """Posts the given graph to the destination of the streamer using the
        given JSON encoder. When ``encoder`` is ``None``, it falls back to the
        default JSON encoder of the streamer in the `encoder` property.

        ``destination`` must be a file-like object or an instance of
        `GephiConnection`.
        """
        encoder = encoder or self.encoder
        for jsonobj in self.iterjsonobj(graph):
            self.send_event(jsonobj, destination, encoder=encoder, flush=False)
        destination.flush()

    def send_event(self, event, destination, encoder=None, flush=True):
        """Sends a single JSON event to the given destination using the given
        JSON encoder.  When ``encoder`` is ``None``, it falls back to the
        default JSON encoder of the streamer in the `encoder` property.

        ``destination`` must be a file-like object or an instance of
        `GephiConnection`.

        The method flushes the destination after sending the event. If you want
        to avoid this (e.g., because you are sending many events), set
        ``flush`` to ``False``.
        """
        encoder = encoder or self.encoder
        destination.write(encoder.encode(event).encode("utf-8"))
        destination.write(b"\r\n")
        if flush:
            destination.flush()



# vim:ts=4 sw=4 sts=4:

import unittest

from igraph import Graph, Vertex, VertexSeq

from .utils import is_pypy

try:
    import numpy as np
except ImportError:
    np = None


class VertexTests(unittest.TestCase):
    def setUp(self):
        self.g = Graph.Full(10)

    def testHash(self):
        data = {}
        n = self.g.vcount()
        for i in range(n):
            code1 = hash(self.g.vs[i])
            code2 = hash(self.g.vs[i])
            self.assertEqual(code1, code2)
            data[self.g.vs[i]] = i

        for i in range(n):
            self.assertEqual(i, data[self.g.vs[i]])

    def testRichCompare(self):
        g2 = Graph.Full(10)
        for i in range(self.g.vcount()):
            for j in range(self.g.vcount()):
                self.assertEqual(i == j, self.g.vs[i] == self.g.vs[j])
                self.assertEqual(i != j, self.g.vs[i] != self.g.vs[j])
                self.assertEqual(i < j, self.g.vs[i] < self.g.vs[j])
                self.assertEqual(i > j, self.g.vs[i] > self.g.vs[j])
                self.assertEqual(i <= j, self.g.vs[i] <= self.g.vs[j])
                self.assertEqual(i >= j, self.g.vs[i] >= self.g.vs[j])
                self.assertFalse(self.g.vs[i] == g2.vs[j])
                self.assertFalse(self.g.vs[i] != g2.vs[j])
                self.assertFalse(self.g.vs[i] < g2.vs[j])
                self.assertFalse(self.g.vs[i] > g2.vs[j])
                self.assertFalse(self.g.vs[i] <= g2.vs[j])
                self.assertFalse(self.g.vs[i] >= g2.vs[j])
                self.assertFalse(self.g.es[i] == self.g.vs[j])

    def testUpdateAttributes(self):
        v = self.g.vs[0]

        v.update_attributes(a=2)
        self.assertEqual(v["a"], 2)

        v.update_attributes([("a", 3), ("b", 4)], c=5, d=6)
        self.assertEqual(v.attributes(), {"a": 3, "b": 4, "c": 5, "d": 6})

        v.update_attributes({"b": 44, "c": 55})
        self.assertEqual(v.attributes(), {"a": 3, "b": 44, "c": 55, "d": 6})

    def testPhantomVertex(self):
        v = self.g.vs[9]
        v.delete()

        # v is now a phantom vertex; try to freak igraph out now :)
        self.assertRaises(ValueError, v.update_attributes, a=2)
        self.assertRaises(ValueError, v.__getitem__, "a")
        self.assertRaises(ValueError, v.__setitem__, "a", 4)
        self.assertRaises(ValueError, v.__delitem__, "a")
        self.assertRaises(ValueError, v.attributes)

    def testIncident(self):
        g = Graph.Famous("petersen")
        g.to_directed()

        method_table = {"all": "all_edges", "in": "in_edges", "out": "out_edges"}

        for i in range(g.vcount()):
            vertex = g.vs[i]
            for mode, method_name in method_table.items():
                method = getattr(vertex, method_name)
                self.assertEqual(
                    g.incident(i, mode=mode),
                    [edge.index for edge in vertex.incident(mode=mode)],
                )
                self.assertEqual(
                    g.incident(i, mode=mode), [edge.index for edge in method()]
                )

    def testNeighbors(self):
        g = Graph.Famous("petersen")
        g.to_directed()

        for i in range(g.vcount()):
            vertex = g.vs[i]
            for mode in "all in out".split():
                self.assertEqual(
                    g.neighbors(i, mode=mode),
                    [edge.index for edge in vertex.neighbors(mode=mode)],
                )

    @unittest.skipIf(
        is_pypy, "skipped on PyPy because we do not have access to docstrings"
    )
    def testProxyMethods(self):
        # We only test with connected graphs because disconnected graphs might
        # print a warning when distances() is invoked on them and we want
        # to avoid that in the test output.
        while True:
            g = Graph.GRG(10, 0.6)
            if g.is_connected():
                break

        v = g.vs[0]

        # - neighbors(), predecessors() and succesors() are ignored because they
        #   return vertex lists while the methods in Graph return vertex index
        #   lists.
        # - incident(), all_edges(), in_edges() and out_edges() are ignored
        #   because it returns an edge list while the methods in Graph return
        #   edge indices.
        # - pagerank() and personalized_pagerank() are ignored because of numerical
        #   inaccuracies
        # - shortest_paths() is ignored because it's a deprecated alias
        # - delete() is ignored because it mutates the graph
        ignore = (
            "neighbors predecessors successors pagerank personalized_pagerank"
            " delete incident all_edges in_edges out_edges shortest_paths"
        )
        ignore = set(ignore.split())

        # Methods not listed here are expected to return an int or a float
        return_types = {"distances": list, "get_shortest_paths": list}

        for name in Vertex.__dict__:
            if name in ignore:
                continue

            func = getattr(v, name)
            docstr = func.__doc__

            if not docstr.startswith("Proxy method"):
                continue

            result = func()
            self.assertEqual(
                getattr(g, name)(v.index),
                result,
                msg=("Vertex.%s proxy method misbehaved" % name),
            )

            return_type = return_types.get(name, (int, float))
            self.assertTrue(
                isinstance(result, return_type),
                msg=("Vertex.%s proxy method did not return %s" % (name, return_type)),
            )


class VertexSeqTests(unittest.TestCase):
    def setUp(self):
        self.g = Graph.Full(10)
        self.g.vs["test"] = list(range(10))
        self.g.vs["name"] = list("ABCDEFGHIJ")

    def testCreation(self):
        self.assertTrue(len(VertexSeq(self.g)) == 10)
        self.assertTrue(len(VertexSeq(self.g, 2)) == 1)
        self.assertTrue(len(VertexSeq(self.g, [1, 2, 3])) == 3)
        self.assertTrue(VertexSeq(self.g, [1, 2, 3]).indices == [1, 2, 3])
        self.assertRaises(ValueError, VertexSeq, self.g, 12)
        self.assertRaises(ValueError, VertexSeq, self.g, [12])
        self.assertTrue(self.g.vs.graph == self.g)

    def testIndexing(self):
        n = self.g.vcount()
        for i in range(n):
            self.assertEqual(i, self.g.vs[i].index)
            self.assertEqual(n - i - 1, self.g.vs[-i - 1].index)
        self.assertRaises(IndexError, self.g.vs.__getitem__, n)
        self.assertRaises(IndexError, self.g.vs.__getitem__, -n - 1)
        self.assertRaises(TypeError, self.g.vs.__getitem__, 1.5)

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testNumPyIndexing(self):
        n = self.g.vcount()
        for i in range(self.g.vcount()):
            arr = np.array([i])
            self.assertEqual(i, self.g.vs[arr[0]].index)
            arr = np.array([-i - 1])
            self.assertEqual(n - i - 1, self.g.vs[arr[0]].index)

        arr = np.array([n])
        self.assertRaises(IndexError, self.g.vs.__getitem__, arr[0])

        arr = np.array([-n - 1])
        self.assertRaises(IndexError, self.g.vs.__getitem__, arr[0])

        arr = np.array([1.5])
        self.assertRaises(TypeError, self.g.vs.__getitem__, arr[0])

        ind = [1, 3, 5, 8, 3, 2]
        arr = np.array(ind)
        self.assertEqual(ind, [vertex.index for vertex in self.g.vs[arr.tolist()]])
        self.assertEqual(ind, [vertex.index for vertex in self.g.vs[list(arr)]])

    def testPartialAttributeAssignment(self):
        only_even = self.g.vs.select(lambda v: (v.index % 2 == 0))
        only_even["test"] = [0] * len(only_even)
        self.assertTrue(self.g.vs["test"] == [0, 1, 0, 3, 0, 5, 0, 7, 0, 9])
        only_even["test2"] = list(range(5))
        self.assertTrue(
            self.g.vs["test2"] == [0, None, 1, None, 2, None, 3, None, 4, None]
        )

    def testSequenceReusing(self):
        if "test" in self.g.vertex_attributes():
            del self.g.vs["test"]

        self.g.vs["test"] = ["A", "B", "C"]
        self.assertTrue(
            self.g.vs["test"] == ["A", "B", "C", "A", "B", "C", "A", "B", "C", "A"]
        )
        self.g.vs["test"] = "ABC"
        self.assertTrue(self.g.vs["test"] == ["ABC"] * 10)

        only_even = self.g.vs.select(lambda v: (v.index % 2 == 0))
        only_even["test"] = ["D", "E"]
        self.assertTrue(
            self.g.vs["test"]
            == ["D", "ABC", "E", "ABC", "D", "ABC", "E", "ABC", "D", "ABC"]
        )
        del self.g.vs["test"]
        only_even["test"] = ["D", "E"]
        self.assertTrue(
            self.g.vs["test"] == ["D", None, "E", None, "D", None, "E", None, "D", None]
        )

    def testAllSequence(self):
        self.assertTrue(len(self.g.vs) == 10)
        self.assertTrue(self.g.vs["test"] == list(range(10)))

    def testEmptySequence(self):
        empty_vs = self.g.vs.select(None)
        self.assertTrue(len(empty_vs) == 0)
        self.assertRaises(IndexError, empty_vs.__getitem__, 0)
        self.assertRaises(KeyError, empty_vs.__getitem__, "nonexistent")
        self.assertTrue(empty_vs["test"] == [])
        empty_vs = self.g.vs[[]]
        self.assertTrue(len(empty_vs) == 0)
        empty_vs = self.g.vs[()]
        self.assertTrue(len(empty_vs) == 0)

    def testCallableFilteringFind(self):
        vertex = self.g.vs.find(lambda v: (v.index % 2 == 1))
        self.assertTrue(vertex.index == 1)
        self.assertRaises(IndexError, self.g.vs.find, lambda v: (v.index % 2 == 3))

    def testCallableFilteringSelect(self):
        only_even = self.g.vs.select(lambda v: (v.index % 2 == 0))
        self.assertTrue(len(only_even) == 5)
        self.assertRaises(KeyError, only_even.__getitem__, "nonexistent")
        self.assertTrue(only_even["test"] == [0, 2, 4, 6, 8])

    def testChainedCallableFilteringSelect(self):
        only_div_six = self.g.vs.select(
            lambda v: (v.index % 2 == 0), lambda v: (v.index % 3 == 0)
        )
        self.assertTrue(len(only_div_six) == 2)
        self.assertTrue(only_div_six["test"] == [0, 6])

        only_div_six = self.g.vs.select(lambda v: (v.index % 2 == 0)).select(
            lambda v: (v.index % 3 == 0)
        )
        self.assertTrue(len(only_div_six) == 2)
        self.assertTrue(only_div_six["test"] == [0, 6])

    def testIntegerFilteringFind(self):
        self.assertEqual(self.g.vs.find(3).index, 3)
        self.assertEqual(self.g.vs.select(2, 3, 4, 2).find(3).index, 2)
        self.assertRaises(IndexError, self.g.vs.find, 17)

    def testIntegerFilteringSelect(self):
        subset = self.g.vs.select(2, 3, 4, 2)
        self.assertEqual(len(subset), 4)
        self.assertEqual(subset["test"], [2, 3, 4, 2])
        self.assertRaises(TypeError, self.g.vs.select, 2, 3, 4, 2, None)

        subset = self.g.vs[2, 3, 4, 2]
        self.assertTrue(len(subset) == 4)
        self.assertTrue(subset["test"] == [2, 3, 4, 2])

    def testStringFilteringFind(self):
        self.assertEqual(self.g.vs.find("D").index, 3)
        self.assertEqual(self.g.vs.select(2, 3, 4, 2).find("C").index, 2)
        self.assertRaises(ValueError, self.g.vs.select(2, 3, 4, 2).find, "F")
        self.assertRaises(ValueError, self.g.vs.find, "NoSuchName")

    def testIterableFilteringSelect(self):
        subset = self.g.vs.select(list(range(5, 8)))
        self.assertTrue(len(subset) == 3)
        self.assertTrue(subset["test"] == [5, 6, 7])

    def testSliceFilteringSelect(self):
        subset = self.g.vs.select(slice(5, 8))
        self.assertTrue(len(subset) == 3)
        self.assertTrue(subset["test"] == [5, 6, 7])
        subset = self.g.vs[5:16:2]
        self.assertTrue(len(subset) == 3)
        self.assertTrue(subset["test"] == [5, 7, 9])

    def testKeywordFilteringSelect(self):
        g = Graph.Barabasi(10000)
        g.vs["degree"] = g.degree()
        g.vs["parity"] = [i % 2 for i in range(g.vcount())]
        length = len(g.vs(degree_gt=30))
        self.assertTrue(length < 1000)
        self.assertTrue(len(g.vs(degree_gt=30, parity=0)) <= 500)
        del g.vs["degree"]
        self.assertTrue(len(g.vs(_degree_gt=30)) == length)

    def testIndexAndKeywordFilteringFind(self):
        self.assertRaises(ValueError, self.g.vs.find, 2, name="G")
        self.assertRaises(ValueError, self.g.vs.find, 2, test=4)
        self.assertTrue(self.g.vs.find(2, name="C") == self.g.vs[2])
        self.assertTrue(self.g.vs.find(2, test=2) == self.g.vs[2])

    def testIndexOutOfBoundsSelect(self):
        g = Graph.Full(3)
        self.assertRaises(ValueError, g.vs.select, 4)
        self.assertRaises(ValueError, g.vs.select, 4, 5)
        self.assertRaises(ValueError, g.vs.select, (4, 5))
        self.assertRaises(ValueError, g.vs.select, 2, -1)
        self.assertRaises(ValueError, g.vs.select, (2, -1))
        self.assertRaises(ValueError, g.vs.__getitem__, (0, 1000000))

    def testGraphMethodProxying(self):
        g = Graph.Barabasi(100)
        vs = g.vs(1, 3, 5, 7, 9)
        self.assertEqual(vs.degree(), g.degree(vs))
        self.assertEqual(g.degree(vs), g.degree(vs.indices))
        for v, d in zip(vs, vs.degree()):
            self.assertEqual(v.degree(), d)

    def testBug73(self):
        # This is a regression test for igraph/python-igraph#73
        g = Graph()
        g.add_vertices(3)
        g.vs[0]["name"] = 1
        g.vs[1]["name"] = "h"
        g.vs[2]["name"] = 17

        self.assertEqual(1, g.vs.find("h").index)
        self.assertEqual(1, g.vs.find(1).index)
        self.assertEqual(0, g.vs.find(name=1).index)
        self.assertEqual(2, g.vs.find(name=17).index)

    def testBug367(self):
        # This is a regression test for igraph/python-igraph#367
        g = Graph()
        g.add_vertices([1, 2, 5])
        self.assertEqual([1, 2, 5], g.vs["name"])
        self.assertEqual(2, g.vs.find(name=5).index)


def suite():
    vertex_suite = unittest.defaultTestLoader.loadTestsFromTestCase(VertexTests)
    vs_suite = unittest.defaultTestLoader.loadTestsFromTestCase(VertexSeqTests)
    return unittest.TestSuite([vertex_suite, vs_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



# vim:ts=4 sw=4 sts=4:
import unittest

from igraph import Graph


class AttributeTests(unittest.TestCase):
    def testGraphAttributes(self):
        g = Graph.Full(5)
        g["date"] = "2005-12-17"
        self.assertTrue(g["date"] == "2005-12-17")
        del g["date"]
        self.assertRaises(KeyError, g.__getitem__, "date")

    def testVertexAttributes(self):
        g = Graph.Full(5)
        g.vs[0]["name"] = "first"
        self.assertTrue(g.vs[0]["name"] == "first")
        del g.vs["name"]
        self.assertRaises(KeyError, g.vs.__getitem__, "name")

        g.vs[0]["name"] = "second"
        g.vs[0]["date"] = "2007-12-17"
        ans = g.vs[0].attribute_names()
        ans.sort()
        self.assertTrue(ans == ["date", "name"])
        attrs = g.vs[0].attributes()
        self.assertTrue(attrs == {"name": "second", "date": "2007-12-17"})

    def testEdgeAttributes(self):
        g = Graph.Full(5)
        g.es[0]["name"] = "first"
        self.assertTrue(g.es[0]["name"] == "first")
        del g.es["name"]
        self.assertRaises(KeyError, g.es.__getitem__, "name")

        g.es[0]["name"] = "second"
        g.es[0]["date"] = "2007-12-17"
        ans = g.es[0].attribute_names()
        ans.sort()
        self.assertTrue(ans == ["date", "name"])
        attrs = g.es[0].attributes()
        self.assertTrue(attrs == {"name": "second", "date": "2007-12-17"})

    def testMassVertexAttributeAssignment(self):
        g = Graph.Full(5)
        g.vs.set_attribute_values("name", list(range(5)))
        self.assertTrue(g.vs.get_attribute_values("name") == list(range(5)))
        g.vs["name"] = list(range(5, 10))
        self.assertTrue(g.vs["name"] == list(range(5, 10)))
        g.vs["name2"] = (1, 2, 3, 4, 6)
        self.assertTrue(g.vs["name2"] == [1, 2, 3, 4, 6])
        g.vs.set_attribute_values("name", [2])
        self.assertTrue(g.vs["name"] == [2] * 5)

    def testMassEdgeAttributeAssignment(self):
        g = Graph.Full(5)
        g.es.set_attribute_values("name", list(range(10)))
        self.assertTrue(g.es.get_attribute_values("name") == list(range(10)))
        g.es["name"] = list(range(10, 20))
        self.assertTrue(g.es["name"] == list(range(10, 20)))
        g.es["name2"] = (1, 2, 3, 4, 6, 1, 2, 3, 4, 6)
        self.assertTrue(g.es["name2"] == [1, 2, 3, 4, 6, 1, 2, 3, 4, 6])
        g.es.set_attribute_values("name", [2])
        self.assertTrue(g.es["name"] == [2] * 10)

    def testVertexNameIndexing(self):
        g = Graph.Famous("bull")
        g.vs["name"] = ["foo", "bar", "baz", "fred", "thud"]
        self.assertTrue(g.degree("bar") == 3)
        self.assertTrue(g.degree(["bar", "fred", 0]) == [3, 1, 2])
        g.vs[2]["name"] = "quack"
        self.assertRaises(ValueError, g.degree, "baz")
        self.assertTrue(g.degree("quack") == 3)
        self.assertTrue(g.degree("quack") == 3)
        self.assertTrue(g.degree(["bar", "thud", 0]) == [3, 1, 2])
        del g.vs["name"]
        self.assertRaises(ValueError, g.degree, ["bar", "thud", 0])

    def testVertexNameIndexingBytes(self):
        g = Graph.Famous("bull")
        g.vs["name"] = [b"foo", b"bar", b"baz", b"fred", b"thud"]
        self.assertTrue(g.degree(b"bar") == 3)
        self.assertTrue(g.degree([b"bar", b"fred", 0]) == [3, 1, 2])
        g.vs[2]["name"] = b"quack"
        self.assertRaises(ValueError, g.degree, b"baz")
        self.assertTrue(g.degree(b"quack") == 3)
        del g.vs["name"]
        self.assertRaises(ValueError, g.degree, [b"bar", b"thud", 0])

    def testUnhashableVertexNames(self):
        g = Graph.Famous("bull")
        g.vs["name"] = [str(x) for x in range(4)]

        value = "this is not hashable".split()
        g.vs[2]["name"] = value

        # Trigger an indexing by doing a lookup by name
        try:
            g.vs.find("3")
            err = None
        except Exception as ex:
            err = ex

        # Check the exception
        self.assertTrue(isinstance(err, RuntimeError))
        self.assertTrue(repr(value) in str(err))

    def testVertexNameIndexingBug196(self):
        g = Graph()
        a, b = b"a", b"b"
        g.add_vertices([a, b])
        g.add_edges([(a, b)])
        self.assertEqual(g.ecount(), 1)
        self.assertTrue(g.are_adjacent(a, b))

    def testInvalidAttributeNames(self):
        g = Graph.Famous("bull")
        for attr_name in [None, 2.654, unittest, str]:
            self.assertRaises(TypeError, g.vs.__setitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.vs.__getitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.vs[0].__setitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.vs[0].__getitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.es.__setitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.es.__getitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.es[0].__setitem__, attr_name, "foo")
            self.assertRaises(TypeError, g.es[0].__getitem__, attr_name, "foo")


class AttributeCombinationTests(unittest.TestCase):
    def setUp(self):
        el = [(0, 1), (1, 0), (1, 2), (2, 3), (2, 3), (2, 3), (3, 3)]
        self.g = Graph(el)
        self.g.es["weight"] = [1, 2, 3, 4, 5, 6, 7]
        self.g.es["weight2"] = [1, 2, 3, 4, 5, 6, 7]

    def testCombinationMax(self):
        g = self.g
        g.simplify(combine_edges="max")
        self.assertTrue(g.es["weight"] == [2, 3, 6])
        self.assertTrue(g.es["weight2"] == [2, 3, 6])

    def testCombinationMin(self):
        g = self.g
        g.simplify(combine_edges="min")
        self.assertTrue(g.es["weight"] == [1, 3, 4])
        self.assertTrue(g.es["weight2"] == [1, 3, 4])

    def testCombinationRandom(self):
        g = self.g
        g.simplify(combine_edges="random")
        del g.es["weight2"]
        for _i in range(100):
            self.assertTrue(g.es[0]["weight"] in (1, 2))
            self.assertTrue(g.es[1]["weight"] == 3)
            self.assertTrue(g.es[2]["weight"] in (4, 5, 6))

    def testCombinationMean(self):
        g = self.g
        g.simplify(combine_edges="mean")
        self.assertTrue(g.es["weight"] == [1.5, 3, 5])
        self.assertTrue(g.es["weight2"] == [1.5, 3, 5])

    def testCombinationMedian(self):
        g = self.g
        g.es["weight2"] = [1, 0, 2, 4, 8, 6, 7]
        g.simplify(combine_edges="median")
        self.assertTrue(g.es["weight"] == [1.5, 3, 5])
        self.assertTrue(g.es["weight2"] == [0.5, 2, 6])

    def testCombinationSum(self):
        g = self.g
        g.simplify(combine_edges="sum")
        self.assertTrue(g.es["weight"] == [3, 3, 15])
        self.assertTrue(g.es["weight2"] == [3, 3, 15])

    def testCombinationProd(self):
        g = self.g
        g.simplify(combine_edges="prod")
        self.assertTrue(g.es["weight"] == [2, 3, 120])
        self.assertTrue(g.es["weight2"] == [2, 3, 120])

    def testCombinationFirst(self):
        g = self.g
        g.es["weight2"] = [1, 0, 2, 6, 8, 4, 7]
        g.simplify(combine_edges="first")
        self.assertTrue(g.es["weight"] == [1, 3, 4])
        self.assertTrue(g.es["weight2"] == [1, 2, 6])

    def testCombinationLast(self):
        g = self.g
        g.es["weight2"] = [1, 0, 2, 6, 8, 4, 7]
        g.simplify(combine_edges="last")
        self.assertTrue(g.es["weight"] == [2, 3, 6])
        self.assertTrue(g.es["weight2"] == [0, 2, 4])

    def testCombinationConcat(self):
        g = self.g
        g.es["name"] = list("ABCDEFG")
        g.simplify(combine_edges={"name": "concat"})
        self.assertFalse("weight" in g.edge_attributes())
        self.assertFalse("weight2" in g.edge_attributes())
        self.assertTrue(g.es["name"] == ["AB", "C", "DEF"])

    def testCombinationMaxMinIgnore(self):
        g = self.g
        g.es["name"] = list("ABCDEFG")
        g.simplify(combine_edges={"weight": "min", "weight2": "max", "name": "ignore"})
        self.assertTrue(g.es["weight"] == [1, 3, 4])
        self.assertTrue(g.es["weight2"] == [2, 3, 6])
        self.assertFalse("name" in g.edge_attributes())

    def testCombinationIgnoreAsNone(self):
        g = self.g
        g.es["name"] = list("ABCDEFG")
        g.simplify(combine_edges={"weight": "min", "name": None})
        self.assertTrue(g.es["weight"] == [1, 3, 4])
        self.assertFalse("weight2" in g.edge_attributes())
        self.assertFalse("name" in g.edge_attributes())

    def testCombinationFunction(self):
        g = self.g

        def join_dash(items):
            return "-".join(items)

        g.es["name"] = list("ABCDEFG")
        g.simplify(combine_edges={"weight": max, "name": join_dash})
        self.assertTrue(g.es["weight"] == [2, 3, 6])
        self.assertFalse("weight2" in g.edge_attributes())
        self.assertTrue(g.es["name"] == ["A-B", "C", "D-E-F"])

    def testCombinationDefault(self):
        g = self.g
        g.simplify(combine_edges={None: "max"})
        self.assertTrue(g.es["weight"] == [2, 3, 6])
        self.assertTrue(g.es["weight2"] == [2, 3, 6])

    def testCombinationDefaultOverride(self):
        g = self.g
        g.simplify(combine_edges={None: "max", "weight": "sum"})
        self.assertTrue(g.es["weight"] == [3, 3, 15])
        self.assertTrue(g.es["weight2"] == [2, 3, 6])

    def testCombinationTypeMismatch(self):
        g = self.g
        g.es["weight"] = list("ABCDEFG")
        self.assertRaises(TypeError, g.simplify, combine_edges={"weight": "mean"})

    def testCombinationNonexistentAttribute(self):
        g = self.g
        g.simplify(combine_edges={"nonexistent": max})
        self.assertTrue(g.edge_attributes() == [])

    def testCombinationNone(self):
        g = self.g
        g.simplify()
        self.assertTrue(sorted(g.edge_attributes()) == [])


class UnicodeAttributeTests(unittest.TestCase):
    def testUnicodeAttributeNameCombination(self):
        g = Graph.Erdos_Renyi(n=9, m=20)
        g.es["test"] = 1
        g.contract_vertices([0, 0, 0, 1, 1, 1, 2, 2, 2])


def suite():
    attribute_suite = unittest.defaultTestLoader.loadTestsFromTestCase(AttributeTests)
    attribute_combination_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        AttributeCombinationTests
    )
    unicode_attributes_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        UnicodeAttributeTests
    )
    return unittest.TestSuite(
        [
            attribute_suite,
            attribute_combination_suite,
            unicode_attributes_suite,
        ]
    )


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()
